"""
"""

from . import _preprocessors, augmenters, components, databases, model_configs, models, preprocessors, utils
from .version import __version__

__all__ = [
    "utils",
    "_preprocessors",
    "preprocessors",
    "augmenters",
    "databases",
    "components",
    "models",
    "model_configs",
    "__version__",
]



"""
"""

from abc import ABC, abstractmethod
from typing import Any, List, Optional, Sequence, Tuple

import numpy as np
import torch.nn as nn
from torch import Tensor

from ..cfg import DEFAULTS
from ..utils.misc import ReprMixin, add_docstring

__all__ = [
    "Augmenter",
]


_augmenter_forward_doc = """Forward method of the augmenter.

        Parameters
        ----------
        sig : torch.Tensor
            Batched ECGs to be augmented, of shape ``(batch, lead, siglen)``.
        label : torch.Tensor, optional
            Batched labels of the ECGs.
        *extra_tensors : Sequence[torch.Tensor], optional
            Extra tensors to be augmented, e.g. masks for custom loss functions, etc.
        **kwargs: dict, optional
            Additional keyword arguments to be passed to the augmenters.

        Returns
        -------
        Sequence[torch.Tensor]
            The augmented ECGs, labels, and optional extra tensors.

        """


class Augmenter(ReprMixin, nn.Module, ABC):
    """Base class for augmenters.

    An Augmentor performs data augmentation on the input ECGs,
    labels, and optional extra tensors.

    """

    __name__ = "Augmentor"

    @add_docstring(_augmenter_forward_doc)
    @abstractmethod
    def forward(
        self, sig: Tensor, label: Optional[Tensor] = None, *extra_tensors: Sequence[Tensor], **kwargs: Any
    ) -> Tuple[Tensor, ...]:
        raise NotImplementedError

    # def __call__(self, sig:Tensor, label:Optional[Tensor]=None, *extra_tensors:Sequence[Tensor], **kwargs:Any) -> Tuple[Tensor, ...]:
    #     """
    #     alias of `self.generate`
    #     """
    #     return self.generate(sig, label, *extra_tensors, **kwargs)

    def get_indices(self, prob: float, pop_size: int, scale_ratio: float = 0.1) -> List[int]:
        """Get a list of indices to be selected.

        A random list of indices in the range ``[0, pop_size-1]``
        is generated, with the probability of each index to be selected.

        Parameters
        ----------
        prob : float
            The probability of each index to be selected.
        pop_size : int
            The size of the population.
        scale_ratio : float, default 0.1
            Scale ratio of std of the normal distribution to the population size.

        Returns
        -------
        indices : List[int],
            A list of indices.

        TODO
        ----
        Add parameter `min_dist` so that
        any 2 selected indices are at least `min_dist` apart.

        """
        k = DEFAULTS.RNG.normal(pop_size * prob, scale_ratio * pop_size)
        # print(pop_size * prob, scale_ratio*pop_size)
        k = int(round(np.clip(k, 0, pop_size)))
        indices = DEFAULTS.RNG_sample(list(range(pop_size)), k).tolist()
        return indices



"""
utilities for downloading and extracting compressed data files

For most PhysioNet databases, the WFDB package already has a method `dl_database`
for downloading the data files.

"""

import collections
import os
import re
import shlex
import shutil
import subprocess
import tarfile
import tempfile
import time
import urllib.parse
import warnings
import zipfile
from pathlib import Path
from typing import Any, Iterable, Literal, Optional, Union

import boto3
import requests
from botocore import UNSIGNED
from botocore.client import Config
from tqdm.auto import tqdm

__all__ = [
    "http_get",
]


PHYSIONET_DB_VERSION_PATTERN = "\\d+\\.\\d+\\.\\d+"


def http_get(
    url: str,
    dst_dir: Union[str, bytes, os.PathLike],
    proxies: Optional[dict] = None,
    extract: Literal[True, False, "auto"] = "auto",
    filename: Optional[str] = None,
) -> Path:
    """Download contents of a URL and save to a file.

    This function is a modified version of the `download_file` function in
    `transformers.file_utils` [1]_.

    Parameters
    ----------
    url : str
        URL to download file from.
    dst_dir : `path-like`
        Directory to save the file.
        If `extract` is ``True``, the extracted content will be saved to `dst_dir`.
    proxies : dict, optional
        Dictionary of proxy settings.
    extract : {True, False, "auto"}, default "auto"
        Whether to extract the downloaded file.
        "auto" will extract if the file is a `zip` file or a compressed `tar` file.
    filename : str, optional
        Name of the downloaded file.
        If None, the filename will be the basename of the URL.
        If `extract` is ``True``, the filename will NOT affect the extracted folder name,
        which is set to `dst_dir`, and `filename` is only the downloaded file name.

        .. versionadded:: 0.0.20

    Returns
    -------
    final_dst : pathlib.Path
        Path to the downloaded file or the extracted folder.

    References
    ----------
    .. [1] https://github.com/huggingface/transformers/blob/master/src/transformers/file_utils.py

    """
    if filename is not None and (Path(dst_dir) / filename).exists():
        raise FileExistsError("file already exists")
    url_parsed = urllib.parse.urlparse(url)
    if url_parsed.scheme == "":
        # assume https by default
        url = f"https://{url}"
        url_parsed = urllib.parse.urlparse(url)
    if url_parsed.scheme not in ["http", "https", "s3"]:
        raise ValueError(f"Unsupported URL scheme {url_parsed.scheme}")

    assert extract in [True, False, "auto"], "extract must be True, False, or 'auto'."

    if url_parsed.scheme == "s3":
        _download_from_aws_s3_using_awscli(url, dst_dir)
        return Path(dst_dir)

    if url_parsed.netloc == "www.dropbox.com" and url_parsed.query == "dl=0":
        url_parsed = url_parsed._replace(query="dl=1")
        url = url_parsed.geturl()
    # for example "https://www.dropbox.com/s/xxx/test%3F.zip??dl=1"
    # produces pure_url = "https://www.dropbox.com/s/xxx/test?.zip"
    pure_url = urllib.parse.unquote(url_parsed._replace(query="").geturl())

    if url_parsed.netloc == "drive.google.com":
        assert filename is not None, "filename can not be inferred from Google Drive URL."
    if extract == "auto":
        # determine whether to extract based on the file extension
        if filename is not None:
            if is_compressed_file(filename):
                extract = True
            else:
                extract = False
        elif is_compressed_file(url_parsed.path):
            extract = True
        else:
            extract = False

    if url_parsed.netloc == "drive.google.com":
        downloaded_file = tempfile.NamedTemporaryFile(
            dir=dst_dir,
            delete=False,
        )
        _download_from_google_drive(url, downloaded_file.name)
        df_suffix = _suffix(filename)
    else:
        print(f"Downloading {url}.")
        if not is_compressed_file(url_parsed.path) and extract:
            if filename is not None:
                if not is_compressed_file(filename):
                    warnings.warn(
                        "filename is given, and it is not a `zip` file or a compressed `tar` file. "
                        "Automatic decompression is turned off.",
                        RuntimeWarning,
                    )
                    extract = False
                else:
                    pass
            else:
                warnings.warn(
                    "URL must be pointing to a `zip` file or a compressed `tar` file. "
                    "Automatic decompression is turned off. "
                    "The user is responsible for decompressing the file manually.",
                    RuntimeWarning,
                )
                extract = False
        parent_dir = Path(dst_dir).parent
        df_suffix = _suffix(pure_url) if filename is None else _suffix(filename)
        downloaded_file = tempfile.NamedTemporaryFile(
            dir=parent_dir,
            suffix=df_suffix,
            delete=False,
        )
        req = requests.get(url, stream=True, proxies=proxies)
        content_length = req.headers.get("Content-Length")
        total = int(content_length) if content_length is not None else None
        if req.status_code in [403, 404]:
            raise Exception(f"Could not reach {url}.")
        progress = tqdm(unit="B", unit_scale=True, total=total, dynamic_ncols=True, mininterval=1.0)
        for chunk in req.iter_content(chunk_size=1024):
            if chunk:  # filter out keep-alive new chunks
                progress.update(len(chunk))
                downloaded_file.write(chunk)
        progress.close()
        downloaded_file.close()

    # add a delay to avoid the error "process cannot access the file because it is being used by another process"
    time.sleep(0.1)

    if extract:
        if ".zip" in df_suffix:
            _unzip_file(str(downloaded_file.name), str(dst_dir))
        # elif ".tar" in df_suffix:  # tar files
        elif re.search("\\.(tar(\\.(gz|bz2|lz|xz|tz|zst))?|tgz|tbz2|tlz|txz|tzst)$", df_suffix) is not None:
            _untar_file(str(downloaded_file.name), str(dst_dir))
        else:
            os.remove(downloaded_file.name)
            raise Exception(f"Unsupported (compressed) archived file type {df_suffix}")
        # avoid the case the compressed file is a folder with the same name
        # DO NOT use _stem(Path(pure_url))
        if filename is None:
            _folder = Path(url).name.replace(_suffix(url), "")
        else:
            _folder = _stem(Path(filename))
        if _folder in os.listdir(dst_dir):
            tmp_folder = str(dst_dir).rstrip(os.sep) + "_tmp"
            # move (rename) the dst_dir to a temporary folder
            os.rename(dst_dir, tmp_folder)
            # move (rename) the extracted folder to the destination folder
            os.rename(Path(tmp_folder) / _folder, dst_dir)
            shutil.rmtree(tmp_folder)
        final_dst = Path(dst_dir)
    else:
        Path(dst_dir).mkdir(parents=True, exist_ok=True)
        if filename is None:
            final_dst = Path(dst_dir) / Path(pure_url).name
        else:
            final_dst = Path(dst_dir) / filename
        shutil.copyfile(downloaded_file.name, final_dst)
    os.remove(downloaded_file.name)
    return final_dst


def _stem(path: Union[str, bytes, os.PathLike]) -> str:
    """Get filename without extension, especially for .tar.xx files.

    Parameters
    ----------
    path : `path-like`
        Path to the file

    Returns
    -------
    str
        Filename without extension.

    """
    ret = Path(path).stem
    if Path(ret).suffix in [".tar", ".gz", ".tz", ".lz", ".bz2", ".xz", ".zip", ".7z"]:
        return _stem(ret)
    return ret


def _suffix(path: Union[str, bytes, os.PathLike], ignore_pattern: str = PHYSIONET_DB_VERSION_PATTERN) -> str:
    """Get file extension, including all suffixes.

    Parameters
    ----------
    path : `path-like`
        Path to the file.
    ignore_pattern : str, optional
        Pattern to ignore in the filename,
        by default `PHYSIONET_DB_VERSION_PATTERN`.

    Returns
    -------
    str
        The full file extension.

    """
    return "".join(Path(re.sub(ignore_pattern, "", str(path))).suffixes)


def is_compressed_file(path: Union[str, bytes, os.PathLike]) -> bool:
    """Check if the path points to a valid compressed file.

    NOTE: a `.pth.tar` file is NOT considered a compressed file.

    Parameters
    ----------
    path : `path-like`
        Path to the file

    Returns
    -------
    bool
        True if the file is a valid compressed file,
        False otherwise.

    """
    # compressed_file_pattern = "(\\.zip)|(\\.tar)"
    compressed_file_pattern = "^(?!.*\\.pth\\.tar$).*\\.(zip|7z|tar(\\.(gz|bz2|lz|xz|tz|zst))?|tgz|tbz2|tlz|txz|tzst)$"
    return re.search(compressed_file_pattern, _suffix(path)) is not None


def _unzip_file(path_to_zip_file: Union[str, bytes, os.PathLike], dst_dir: Union[str, bytes, os.PathLike]) -> None:
    """Unzips a .zip file to folder path.

    Parameters
    ----------
    path_to_zip_file : `path-like`
        Path to the .zip file.
    dst_dir : `path-like`
        Path to the destination folder.

    Returns
    -------
    None

    """
    print(f"Extracting file {path_to_zip_file} to {dst_dir}.")
    with zipfile.ZipFile(str(path_to_zip_file)) as zip_ref:
        zip_ref.extractall(str(dst_dir))


def _untar_file(path_to_tar_file: Union[str, bytes, os.PathLike], dst_dir: Union[str, bytes, os.PathLike]) -> None:
    """Decompress a .tar.xx file to folder path.

    Parameters
    ----------
    path_to_tar_file : `path-like`
        Path to the .tar.xx file.
    dst_dir : `path-like`
        Path to the destination folder.

    Returns
    -------
    None

    """
    print(f"Extracting file {path_to_tar_file} to {dst_dir}.")
    mode = Path(path_to_tar_file).suffix.replace(".", "r:").replace("tar", "").strip(":")
    with tarfile.open(str(path_to_tar_file), mode) as tar_ref:
        # tar_ref.extractall(str(dst_dir))
        # CVE-2007-4559 (related to  CVE-2001-1267):
        # directory traversal vulnerability in `extract` and `extractall` in `tarfile` module
        _safe_tar_extract(tar_ref, str(dst_dir))


def _is_within_directory(directory: Union[str, bytes, os.PathLike], target: Union[str, bytes, os.PathLike]) -> bool:
    """Check if the target is within the directory.

    Parameters
    ----------
    directory : `path-like`
        Directory to check.
    target : `path-like`
        Path to the target.

    Returns
    -------
    bool
        True if the target is within the directory,
        False otherwise.

    """
    abs_directory = os.path.abspath(directory)
    abs_target = os.path.abspath(target)

    prefix = os.path.commonprefix([abs_directory, abs_target])

    return prefix == abs_directory


def _safe_tar_extract(
    tar: tarfile.TarFile,
    dst_dir: Union[str, bytes, os.PathLike],
    members: Optional[Iterable[tarfile.TarInfo]] = None,
    *,
    numeric_owner: bool = False,
) -> None:
    """Extract members from a tarfile **safely** to a destination directory.

    This function prevents path traversal attacks, by checking that the
    extracted files are within the destination directory.

    Parameters
    ----------
    tar : tarfile.TarFile
        The tarfile to extract from.
    dst_dir : `path-like`
        The destination directory
    members : Iterable[tarfile.TarInfo], optional
        The members to extract
        If is None, extract all members,
        otherwise, must be a subset of the list
        returned by :func:`tar.getmembers`.
    numeric_owner : bool, default False
        If True, only the numbers for user/group names are used
        and not the names. For more information,
        see :func:`tarfile.TarFile.extractall`.

    Returns
    -------
    None

    """
    for member in members or tar.getmembers():
        member_path = os.path.join(dst_dir, member.name)
        if not _is_within_directory(dst_dir, member_path):
            raise Exception("Attempted Path Traversal in Tar File")

    tar.extractall(dst_dir, members, numeric_owner=numeric_owner)


def url_is_reachable(url: str, **kwargs: Any) -> bool:
    """Check if a URL is reachable.

    Parameters
    ----------
    url : str
        The URL.
    **kwargs : dict, optional
        Additional keyword arguments to pass to :meth:`requests.head`.

    Returns
    -------
    bool
        Whether the URL is reachable.

    """
    try:
        timeout = kwargs.pop("timeout", 3)
        r = requests.head(url, timeout=timeout, **kwargs)
        # successful responses and redirection messages
        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Status#information_responses
        # Informational responses (100 – 199)
        # Successful responses (200 – 299)
        # Redirection messages (300 – 399)
        # Client error responses (400 – 499)
        # Server error responses (500 – 599)
        return 100 <= r.status_code < 400
    except Exception:
        return False


def _download_from_google_drive(url_or_id: str, output: Union[str, bytes, os.PathLike], quiet: bool = False) -> None:
    """Download a file from Google Drive.

    Parameters
    ----------
    url_or_id : str
        The URL of the file or the file ID.
    output : `path-like`
        The output file path.
    quiet : bool, default False
        Whether to suppress the output.

    Returns
    -------
    None

    """
    try:
        import gdown
    except (ImportError, ModuleNotFoundError):
        raise ImportError("gdown is required to download from Google Drive.")
    file_id_pattern = "^1[A-Za-z0-9-_]{32}$"
    if re.match(file_id_pattern, url_or_id) is not None:
        url_or_id = f"https://drive.google.com/u/0/uc?id={url_or_id}"
    url_parsed = urllib.parse.urlparse(url_or_id)
    if url_parsed.scheme == "" and url_parsed.netloc == "":
        # no scheme is given, e.g. drive.google.com/file/d/xxx/view?usp=sharing
        url_or_id = f"https://{url_or_id}"
        url_parsed = urllib.parse.urlparse(url_or_id)
    # remove trailing query string
    url_or_id = re.sub("/view\\?.*$", "", url_or_id)
    if re.match("^https://drive\\.google.com/file/d/", url_or_id) is not None:
        url_or_id = re.sub(
            "^https://drive\\.google\\.com/file/d/",
            "https://drive.google.com/u/0/uc?id=",
            url_or_id,
        )
        print(f"Redirecting to {url_or_id}")
    gdown.download(url_or_id, str(output), quiet=quiet)


def count_aws_s3_bucket(bucket_name: str, prefix: str = "") -> int:
    """Count the number of objects in an AWS S3 bucket.

    Parameters
    ----------
    bucket_name : str
        The bucket name.
    prefix : str, optional
        The prefix to filter the objects,
        for example, "ludb/1.0.1/".

    Returns
    -------
    int
        The number of objects in the bucket.

    """
    s3 = boto3.client("s3", config=Config(signature_version=UNSIGNED))
    paginator = s3.get_paginator("list_objects_v2")
    page_iterator = paginator.paginate(Bucket=bucket_name, Prefix=prefix)

    object_count = 0
    for page in page_iterator:
        if "Contents" in page:
            object_count += len(page["Contents"])

    s3.close()

    return object_count


def _download_from_aws_s3_using_boto3(url: str, dst_dir: Union[str, bytes, os.PathLike]) -> None:
    """Download a file from AWS S3 using boto3.

    Currently, much slower than using awscli.

    Parameters
    ----------
    url : str
        The URL of the file.
        For example, "s3://bucket-name/files/pre-fix".
    dst_dir : `path-like`
        The output directory.

    Returns
    -------
    None

    """
    pattern = "^s3://(?P<bucket_name>[^/]+)/(?P<prefix>.+)$"
    match = re.match(pattern, url)
    if match is None:
        raise ValueError(f"Invalid S3 URL: {url}")
    bucket_name = match.group("bucket_name")
    prefix = match.group("prefix")
    if prefix.startswith("files/"):
        prefix = prefix.replace("files/", "")

    s3 = boto3.client("s3", config=Config(signature_version=UNSIGNED))
    paginator = s3.get_paginator("list_objects_v2")
    page_iterator = paginator.paginate(Bucket=bucket_name, Prefix=prefix)
    object_count = 0
    for page in page_iterator:
        if "Contents" in page:
            object_count += len(page["Contents"])
    if object_count == 0:
        raise ValueError(f"No objects found in S3 bucket: {bucket_name}")
    print(f"Downloading from S3 bucket: {bucket_name}, prefix: {prefix}, total files: {object_count}")

    with tqdm(total=object_count, dynamic_ncols=True, mininterval=1.0) as pbar:
        for page in page_iterator:
            if "Contents" in page:
                for obj in page["Contents"]:
                    key = obj["Key"]
                    dst_file = Path(dst_dir) / key
                    if not dst_file.parent.exists():
                        dst_file.parent.mkdir(parents=True, exist_ok=True)
                        s3.download_file(bucket_name, key, str(dst_file))
                    pbar.update(1)

    s3.close()


def _download_from_aws_s3_using_awscli(url: str, dst_dir: Union[str, bytes, os.PathLike]) -> None:
    """Download a file from AWS S3 using awscli.

    Parameters
    ----------
    url : str
        The URL of the file.
        For example, "s3://bucket-name/files/pre-fix".
    dst_dir : `path-like`
        The output directory.

    Returns
    -------
    None

    """
    assert shutil.which("aws") is not None, "AWS cli is required to download from S3."

    pattern = "^s3://(?P<bucket_name>[^/]+)/(?P<prefix>.+)$"
    match = re.match(pattern, url)
    if match is None:
        raise ValueError(f"Invalid S3 URL: {url}")
    bucket_name = match.group("bucket_name")
    prefix = match.group("prefix")
    if prefix.startswith("files/"):
        prefix = prefix.replace("files/", "")
    object_count = count_aws_s3_bucket(bucket_name, prefix)
    print(f"Downloading from S3 bucket: {bucket_name}, prefix: {prefix}, total files: {object_count}")

    pbar = tqdm(total=object_count, dynamic_ncols=True, mininterval=1.0)
    download_count = 0
    command = f"aws s3 sync --no-sign-request {url} {dst_dir}"
    process = subprocess.Popen(shlex.split(command), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    debug_stdout = collections.deque(maxlen=10)
    while 1:
        line = process.stdout.readline().decode("utf-8", errors="replace")
        if line.rstrip():
            debug_stdout.append(line)
            if "download: s3:" in line:
                download_count += 1
                pbar.update(1)
        exitcode = process.poll()
        if exitcode is not None:
            for line in process.stdout:
                debug_stdout.append(line.decode("utf-8", errors="replace"))
            if exitcode is not None and exitcode != 0:
                error_msg = "\n".join(debug_stdout)
                process.communicate()
                process.stdout.close()
                raise subprocess.CalledProcessError(exitcode, error_msg)
            else:
                break
    process.communicate()
    process.stdout.close()
    # object_count - download_count files skipped for they already exist
    pbar.update(object_count - download_count)
    pbar.close()
    print(f"Downloaded {download_count} files from S3.")



"""
"""

from pathlib import Path

import pandas as pd
import wfdb

from ...utils.misc import timeout
from . import aha, cinc2020_aux_data, cinc2021_aux_data

__all__ = [
    "cinc2020_aux_data",
    "cinc2021_aux_data",
    "get_physionet_dbs",
    "aha",
]


def get_physionet_dbs(local: bool = True) -> pd.DataFrame:
    """
    load the list of PhysioNet databases,
    locally stored in the file "./physionet_dbs.csv.gz"
    """
    if not local:
        try:
            with timeout(2):
                return (
                    pd.DataFrame(wfdb.get_dbs(), columns=["db_name", "db_description"]).drop_duplicates().reset_index(drop=True)
                )
        except Exception:  # TimeoutError, ConnectionError, ReadTimeout, ...
            print("Failed to fetch the list of PhysioNet databases from the remote server.")
    return pd.read_csv(Path(__file__).absolute().parent / "physionet_dbs.csv.gz")



# -*- coding: utf-8 -*-

import json
import math
import os
from copy import deepcopy
from numbers import Real
from typing import Any, Dict, List, Literal, Optional, Sequence, Tuple, Union

import numpy as np
import wfdb

from ...cfg import CFG
from ...utils.misc import add_docstring
from ...utils.utils_interval import generalized_intervals_intersection
from ..base import DEFAULT_FIG_SIZE_PER_SEC, BeatAnn, DataBaseInfo, PhysioNetDataBase

__all__ = [
    "LTAFDB",
]


_LTAFDB_INFO = DataBaseInfo(
    title="""
    Long Term AF Database
    """,
    about="""
    1. contains 84 long-term ECG recordings of subjects with paroxysmal or sustained atrial fibrillation
    2. each record contains two simultaneously recorded ECG signals digitized at 128 Hz
    3. records have duration 24 - 25 hours
    4. qrs annotations (.qrs files) were produced by an automated QRS detector, in which detected beats (including occasional ventricular ectopic beats) are labelled "N", detected artifacts are labelled "|", and AF terminations are labelled "T" (inserted manually)
    5. atr annotations (.atr files) were obtained by manual review of the output of an automated ECG analysis system; in these annotation files, all detected beats are labelled by type ('"', "+", "A", "N", "Q", "V"), and rhythm changes ("\x01 Aux", "(AB", "(AFIB", "(B", "(IVR", "(N", "(SBR", "(SVTA", "(T", "(VT", "M", "MB", "MISSB", "PSE") are also annotated
    6. Webpage of the database on PhysioNet [1]_. Paper describing the database [2]_.
    """,
    note="""
    1. both channels of the signals have name "ECG"
    2. the automatically generated qrs annotations (.qrs files) contains NO rhythm annotations
    3. `aux_note` of .atr files of all but one ("64") record start with valid rhythms, all but one end with "" ("30" ends with "\x01 Aux")
    4. for more statistics on the whole database, see [ref 3](#ref3)
    """,
    usage=[
        "Atrial fibrillation (AF) detection",
        "(3 or 4) beat type classification",
        "Rhythm classification",
    ],
    references=[
        "https://physionet.org/content/ltafdb/1.0.0/",
        "Petrutiu S, Sahakian AV, Swiryn S. Abrupt changes in fibrillatory wave characteristics at the termination of paroxysmal atrial fibrillation in humans. Europace 9:466-470 (2007).",
    ],
    doi=[
        "10.1093/europace/eum096",
        "10.13026/C2QG6Q",
    ],
)


@add_docstring(_LTAFDB_INFO.format_database_docstring(), mode="prepend")
class LTAFDB(PhysioNetDataBase):
    """
    Parameters
    ----------
    db_dir : `path-like`, optional
        Storage path of the database.
        If not specified, data will be fetched from Physionet.
    working_dir : `path-like`, optional
        Working directory, to store intermediate files and log files.
    verbose : int, default 1
        Level of logging verbosity.
    kwargs : dict, optional
        Auxilliary key word arguments.

    """

    __name__ = "LTAFDB"

    def __init__(
        self,
        db_dir: Optional[Union[str, bytes, os.PathLike]] = None,
        working_dir: Optional[Union[str, bytes, os.PathLike]] = None,
        verbose: int = 1,
        **kwargs: Any,
    ) -> None:
        from matplotlib.pyplot import cm

        super().__init__(
            db_name="ltafdb",
            db_dir=db_dir,
            working_dir=working_dir,
            verbose=verbose,
            **kwargs,
        )
        self.fs = 128
        self.data_ext = "dat"
        self.auto_ann_ext = "qrs"
        self.manual_ann_ext = "atr"
        self.all_leads = [0, 1]
        self._ls_rec()

        self.rhythm_types = [
            "(N",
            "(AB",
            "(AFIB",
            "(B",
            "(IVR",
            "(SBR",
            "(SVTA",
            "(T",
            "(VT",
            "NOISE",  # additional, since head of each record are noisy
        ]  # others include "\x01 Aux", "M", "MB", "MISSB", "PSE"
        self.rhythm_types_map = CFG({k.replace("(", ""): idx for idx, k in enumerate(self.rhythm_types)})
        self.palette = kwargs.get("palette", None)
        if self.palette is None:
            n_colors = len([k for k in self.rhythm_types_map.keys() if k not in ["N", "NOISE"]])
            colors = iter(cm.rainbow(np.linspace(0, 1, n_colors)))
            self.palette = CFG()
            for k in self.rhythm_types_map.keys():
                if k in ["N", "NOISE"]:
                    continue
                self.palette[k] = next(colors)

        self.beat_types = [
            "A",
            "N",
            "Q",
            "V",
            # '"', "+", are not beat types
        ]
        self.palette["qrs"] = "green"

    def get_subject_id(self, rec: Union[str, int]) -> int:
        """Attach a unique subject ID for the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        int
            Subject ID associated with the record.

        """
        raise NotImplementedError

    @add_docstring(
        PhysioNetDataBase.load_data.__doc__.replace(
            "leads: str or int or sequence of str or int, optional,",
            "leads: int or list of int, optional,",
        ).replace("the leads to load", "the lead number(s) to load")
    )
    def load_data(
        self,
        rec: Union[str, int],
        leads: Optional[Union[int, List[int]]] = None,
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        data_format: str = "channel_first",
        units: str = "mV",
        fs: Optional[Real] = None,
        return_fs: bool = False,
    ) -> Union[np.ndarray, Tuple[np.ndarray, Real]]:
        return super().load_data(rec, leads, sampfrom, sampto, data_format, units, fs, return_fs)

    def load_ann(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        rhythm_format: Literal["intervals", "mask"] = "intervals",
        beat_format: Literal["beat", "dict"] = "beat",
        keep_original: bool = False,
    ) -> dict:
        """Load rhythm and beat annotations of the record.

        Rhythm and beat annotations are stored in the
        `aux_note`, `symbol` attributes of corresponding annotation files.
        NOTE that qrs annotations (.qrs files) do NOT contain any rhythm annotations.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        sampfrom : int, optional
            Start index of the annotations to be loaded.
        sampto : int, optional
            End index of the annotations to be loaded.
        rhythm_format : {"intervals", "mask"}, default "intervals"
            Format of returned annotation, case insensitive.
        beat_format : {"beat", "dict"}, default "beat"
            Format of returned annotation, case insensitive.
        keep_original : bool, default False
            If True, indices will keep the same with the annotation file,
            otherwise subtract `sampfrom` if specified.

        Returns
        -------
        ann : dict
            The annotations of ``rhythm`` and ``beat``, with
            ``rhythm`` annotatoins in the format of intervals, or mask;
            ``beat`` annotations in the format of dict or
            :class:`~torch_ecg.databases.BeatAnn`.

        NOTE
        ----
        At head and tail of the record, segments named "NOISE" were added.

        """
        if isinstance(rec, int):
            rec = self[rec]
        ann = {
            "beat": self.load_beat_ann(
                rec,
                sampfrom,
                sampto,
                beat_format,
                keep_original,
            ),
            "rhythm": self.load_rhythm_ann(
                rec,
                sampfrom,
                sampto,
                rhythm_format,
                keep_original,
            ),
        }
        return ann

    def load_rhythm_ann(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        rhythm_format: Literal["intervals", "mask"] = "intervals",
        keep_original: bool = False,
    ) -> Union[Dict[str, list], np.ndarray]:
        """Load rhythm annotations of the record.

        Rhythm annotations are stored in the `aux_note` attribute
        of corresponding annotation files.
        NOTE that qrs annotations (.qrs files) do NOT contain any rhythm annotations.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        sampfrom : int, optional
            Start index of the annotations to be loaded.
        sampto : int, optional
            End index of the annotations to be loaded.
        rhythm_format : {"intervals", "mask"}, default "intervals"
            Format of returned annotation, case insensitive.
        keep_original : bool, default False
            If True, indices will keep the same with the annotation file,
            otherwise subtract `sampfrom` if specified.

        Returns
        -------
        ann : dict or numpy.ndarray
            Annotations in the format of intervals or mask.

        NOTE
        ----
        At head and tail of the record, segments named "NOISE" were added.

        """
        if isinstance(rec, int):
            rec = self[rec]
        assert rhythm_format.lower() in [
            "intervals",
            "mask",
        ], f"rhythm_format must be 'intervals' or 'mask', got {rhythm_format}"
        fp = str(self.get_absolute_path(rec))
        header = wfdb.rdheader(str(fp))
        sig_len = header.sig_len
        sf = sampfrom or 0
        st = sampto or sig_len
        assert st > sf, "`sampto` should be greater than `sampfrom`!"

        simplified_fp = self.db_dir / f"{rec}_ann.json"
        if simplified_fp.is_file():
            ann = CFG(json.loads(simplified_fp.read_text()))
        else:
            wfdb_ann = wfdb.rdann(str(fp), extension=self.manual_ann_ext)

            ann = CFG({k: [] for k in self.rhythm_types_map.keys()})
            critical_points = wfdb_ann.sample.tolist()
            aux_note = wfdb_ann.aux_note
            start = 0
            current_rhythm = "NOISE"
            for idx, rhythm in zip(critical_points, aux_note):
                if rhythm not in self.rhythm_types:
                    continue
                ann[current_rhythm].append([start, idx])
                current_rhythm = rhythm.replace("(", "")
                start = idx
            # all but one end with "" ("30" ends with "\x01 Aux")
            # i.e. none ends with (start of) valid rhythm
            ann[current_rhythm].append([start, critical_points[-1]])
            ann["NOISE"].append([critical_points[-1], sig_len])

            simplified_fp.write_text(json.dumps(ann, ensure_ascii=False))

        ann = CFG({k: generalized_intervals_intersection(l_itv, [[sf, st]]) for k, l_itv in ann.items()})
        ann = CFG({k: l_itv for k, l_itv in ann.items() if len(l_itv) > 0})
        if rhythm_format.lower() == "mask":
            tmp = deepcopy(ann)
            ann = np.full(shape=(st - sf,), fill_value=self.rhythm_types_map.N, dtype=int)
            for rhythm, l_itv in tmp.items():
                for itv in l_itv:
                    ann[itv[0] - sf : itv[1] - sf] = self.rhythm_types_map[rhythm]
        elif not keep_original:
            for k, l_itv in ann.items():
                ann[k] = [[itv[0] - sf, itv[1] - sf] for itv in l_itv]

        return ann

    def load_beat_ann(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        beat_format: Literal["beat", "dict"] = "beat",
        keep_original: bool = False,
    ) -> Union[Dict[str, np.ndarray], List[BeatAnn]]:
        """Load beat annotations of the record.

        Beat annotations are stored in the `symbol` attribute
        of corresponding annotation files.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        sampfrom : int, optional
            Start index of the annotations to be loaded.
        sampto : int, optional
            End index of the annotations to be loaded.
        beat_format : {"beat", "dict"}, default "beat"
            Format of returned annotation, case insensitive.
        keep_original : bool, default False
            If True, indices will keep the same with the annotation file,
            otherwise subtract `sampfrom` if specified.

        Returns
        -------
        ann : dict or list
            Locations (indices) of the all the
            beat types ("A", "N", "Q", "V").

        """
        if isinstance(rec, int):
            rec = self[rec]
        assert beat_format.lower() in [
            "beat",
            "dict",
        ], f"beat_format must be 'beat' or 'dict', got {beat_format}"
        fp = self.get_absolute_path(rec)
        header = wfdb.rdheader(str(fp))
        sig_len = header.sig_len
        sf = sampfrom or 0
        st = sampto or sig_len
        assert st > sf, "`sampto` should be greater than `sampfrom`!"

        wfdb_ann = wfdb.rdann(
            str(fp),
            extension=self.manual_ann_ext,
            sampfrom=sf,
            sampto=sampto,
        )
        ann = CFG({k: [] for k in self.beat_types})
        for idx, bt in zip(wfdb_ann.sample, wfdb_ann.symbol):
            if bt not in self.beat_types:
                continue
            ann[bt].append(idx)
        if not keep_original and sampfrom is not None:
            ann = CFG({k: np.array(v, dtype=int) - sf for k, v in ann.items()})
        else:
            ann = CFG({k: np.array(v, dtype=int) for k, v in ann.items()})

        if beat_format.lower() == "beat":
            ann = [BeatAnn(i, s) for s, l in ann.items() for i in l]
        return ann

    def load_rpeak_indices(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        use_manual: bool = True,
        keep_original: bool = False,
    ) -> np.ndarray:
        """Load rpeak indices of the record.

        Rpeak indices, or equivalently qrs complex locations,
        are stored in the `symbol` attribute of corresponding annotation files,
        regardless of their beat types.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        sampfrom : int, optional
            Start index of the annotations to be loaded.
        sampto : int, optional
            End index of the annotations to be loaded.
        use_manual : bool, default True
            If True, manually annotated beat annotations (qrs) will be used,
            instead of those generated by algorithms.
        keep_original : bool, default False
            If True, indices will keep the same with the annotation file,
            otherwise subtract `sampfrom` if specified.

        Returns
        -------
        rpeak_inds : numpy.ndarray
            Locations (indices) of the all the rpeaks (qrs complexes).

        """
        fp = str(self.get_absolute_path(rec))
        if use_manual:
            ext = self.manual_ann_ext
        else:
            ext = self.auto_ann_ext
        wfdb_ann = wfdb.rdann(
            str(fp),
            extension=ext,
            sampfrom=sampfrom or 0,
            sampto=sampto,
        )
        rpeak_inds = wfdb_ann.sample[np.isin(wfdb_ann.symbol, self.beat_types)]
        if not keep_original and sampfrom is not None:
            rpeak_inds = rpeak_inds - sampfrom
        return rpeak_inds

    def plot(
        self,
        rec: Union[str, int],
        data: Optional[np.ndarray] = None,
        ann: Optional[Dict[str, np.ndarray]] = None,
        beat_ann: Optional[Dict[str, np.ndarray]] = None,
        rpeak_inds: Optional[Union[Sequence[int], np.ndarray]] = None,
        ticks_granularity: int = 0,
        leads: Optional[Union[int, List[int]]] = None,
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        same_range: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        Plot the signals of a record or external signals (units in μV),
        with metadata (fs, labels, tranche, etc.),
        possibly also along with wave delineations.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        data : numpy.ndarray, optional
            (2-lead) ECG signal to plot,
            should be of the format "channel_first",
            and compatible with `leads`.
            If is not None, data of `rec` will not be used.
            This is useful when plotting filtered data.
        ann : dict, optional
            Rhythm annotations for `data`, covering those from annotation files,
            in the form of ``{k: l_itv, ...}``,
            where ``k`` are listed in `self.rhythm_types_map`,
            and ``l_itv`` are of the form of ``[[a, b], ...]``.
            Ignored if `data` is None
        beat_ann : dict, optional
            Beat annotations for `data`, covering those from annotation files,
            in the form of ``{k: l_inds, ...}``,
            where ``k`` are listed in `self.beat_types`,
            and `l_inds` are array of indices.
            Ignored if `data` is None.
        rpeak_inds : array_like, optional
            Indices of R peaks, covering those from annotation files.
            If `data` is None, then indices should be
            absolute indices in the record
        ticks_granularity : int, default 0
            Granularity to plot axis ticks, the higher the more ticks.
            0 (no ticks) --> 1 (major ticks) --> 2 (major + minor ticks)
        leads : int or List[int], optional,
            The lead number(s) of the data to plot.
        sampfrom : int, optional
            Start index of the data to plot.
        sampto : int, optional
            End index of the data to plot.
        same_range : bool, default False
            If True, all leads are forced to have the same y range.
        kwargs : dict, optional
            Additional arguments to be passed to `matplotlib.pyplot.plot`, etc.

        """
        if isinstance(rec, int):
            rec = self[rec]
        if "plt" not in dir():
            import matplotlib.pyplot as plt

            plt.MultipleLocator.MAXTICKS = 3000
        if leads is None or leads == "all":
            _leads = self.all_leads
        elif isinstance(leads, int):
            _leads = [leads]
        else:
            _leads = leads
        assert all([ld in self.all_leads for ld in _leads])

        lead_indices = [self.all_leads.index(ld) for ld in _leads]
        if data is None:
            _data = self.load_data(
                rec,
                leads=_leads,
                sampfrom=sampfrom,
                sampto=sampto,
                data_format="channel_first",
                units="μV",
            )
        else:
            units = self._auto_infer_units(data)
            self.logger.info(f"input data is auto detected to have units in {units}")
            if units.lower() == "mv":
                _data = 1000 * data
            else:
                _data = data
            _leads = list(range(_data.shape[0]))
        if ann is None and data is None:
            _ann = self.load_rhythm_ann(
                rec,
                sampfrom=sampfrom,
                sampto=sampto,
                rhythm_format="intervals",
                keep_original=False,
            )
        else:
            _ann = ann or CFG({k: [] for k in self.rhythm_types_map.keys()})
        # indices to time
        _ann = {k: [[itv[0] / self.fs, itv[1] / self.fs] for itv in l_itv] for k, l_itv in _ann.items()}
        if rpeak_inds is None and data is None:
            _rpeak = self.load_rpeak_indices(
                rec,
                sampfrom=sampfrom,
                sampto=sampto,
                use_manual=True,
                keep_original=False,
            )
            _rpeak = _rpeak / self.fs  # indices to time
        else:
            _rpeak = np.array(rpeak_inds or []) / self.fs  # indices to time
        if beat_ann is None and data is None:
            _beat_ann = self.load_beat_ann(
                rec,
                beat_format="dict",
                sampfrom=sampfrom,
                sampto=sampto,
                keep_original=False,
            )
        else:
            _beat_ann = beat_ann or CFG({k: [] for k in self.beat_types})
        _beat_ann = {k: [i / self.fs for i in l_inds] for k, l_inds in _beat_ann.items()}  # indices to time

        ann_plot_alpha = 0.2
        rpeaks_plot_alpha = 0.8

        nb_leads = len(_leads)

        line_len = self.fs * 25  # 25 seconds
        nb_lines = math.ceil(_data.shape[1] / line_len)

        for seg_idx in range(nb_lines):
            seg_data = _data[..., seg_idx * line_len : (seg_idx + 1) * line_len]
            secs = (np.arange(seg_data.shape[1]) + seg_idx * line_len) / self.fs
            seg_ann = {k: generalized_intervals_intersection(l_itv, [[secs[0], secs[-1]]]) for k, l_itv in _ann.items()}
            seg_rpeaks = _rpeak[np.where((_rpeak >= secs[0]) & (_rpeak < secs[-1]))[0]]
            seg_beat_ann = {k: [i for i in l_inds if secs[0] <= i <= secs[-1]] for k, l_inds in _beat_ann.items()}
            fig_sz_w = int(round(DEFAULT_FIG_SIZE_PER_SEC * seg_data.shape[1] / self.fs))
            if same_range:
                y_ranges = np.ones((seg_data.shape[0],)) * np.max(np.abs(seg_data)) + 100
            else:
                y_ranges = np.max(np.abs(seg_data), axis=1) + 100
            fig_sz_h = 6 * y_ranges / 1500
            fig, axes = plt.subplots(nb_leads, 1, sharex=True, figsize=(fig_sz_w, np.sum(fig_sz_h)))
            if nb_leads == 1:
                axes = [axes]
            for idx in range(nb_leads):
                axes[idx].plot(secs, seg_data[idx], color="black", label=f"lead - {_leads[idx]}")
                axes[idx].axhline(y=0, linestyle="-", linewidth="1.0", color="red")
                # NOTE that `Locator` has default `MAXTICKS` equal to 1000
                if ticks_granularity >= 1:
                    axes[idx].xaxis.set_major_locator(plt.MultipleLocator(0.2))
                    axes[idx].yaxis.set_major_locator(plt.MultipleLocator(500))
                    axes[idx].grid(which="major", linestyle="-", linewidth="0.5", color="red")
                if ticks_granularity >= 2:
                    axes[idx].xaxis.set_minor_locator(plt.MultipleLocator(0.04))
                    axes[idx].yaxis.set_minor_locator(plt.MultipleLocator(100))
                    axes[idx].grid(which="minor", linestyle=":", linewidth="0.5", color="black")
                for k, l_itv in seg_ann.items():
                    if k in ["N", "NOISE"]:
                        continue
                    for itv in l_itv:
                        axes[idx].axvspan(
                            itv[0],
                            itv[1],
                            color=self.palette[k],
                            alpha=ann_plot_alpha,
                            label=k,
                        )
                for ri in seg_rpeaks:
                    axes[idx].axvspan(
                        ri - 0.01,
                        ri + 0.01,
                        color=self.palette["qrs"],
                        alpha=rpeaks_plot_alpha,
                    )
                for k, l_t in seg_beat_ann.items():
                    for t in l_t:
                        x_pos = t + 0.05 if t + 0.05 < secs[-1] else t - 0.15
                        axes[idx].text(x_pos, 0.65 * y_ranges[idx], k, color="black", fontsize=16)
                axes[idx].legend(loc="upper left")
                axes[idx].set_xlim(secs[0], secs[-1])
                axes[idx].set_ylim(-y_ranges[idx], y_ranges[idx])
                axes[idx].set_xlabel("Time [s]")
                axes[idx].set_ylabel("Voltage [μV]")
            plt.subplots_adjust(hspace=0.2)
            plt.show()

    @property
    def database_info(self) -> DataBaseInfo:
        return _LTAFDB_INFO



"""
"""

import os
import random
import re
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import Any, MutableMapping, Optional, Union

import numpy as np
import torch

__all__ = [
    "CFG",
    "DEFAULTS",
]


_PACKAGE_CACHE = Path("~").expanduser() / ".cache" / "torch_ecg"
if os.access(_PACKAGE_CACHE.parent, os.W_OK) and not _PACKAGE_CACHE.exists():
    _PACKAGE_CACHE.mkdir(parents=True, exist_ok=True)
_DATA_CACHE = _PACKAGE_CACHE / "data"
if os.access(_DATA_CACHE.parent, os.W_OK) and not _DATA_CACHE.exists():
    _DATA_CACHE.mkdir(parents=True, exist_ok=True)


class CFG(dict):
    """This class is created in order to renew the :meth:`update` method,
    to fit the hierarchical structure of configurations.

    Examples
    --------
    >>> c = CFG(hehe={"a": 1, "b": 2})
    >>> c.update(hehe={"a": [-1]})
    >>> c
    {'hehe': {'a': [-1], 'b': 2}}
    >>> c.update(hehe={"c": -10})
    >>> c
    {'hehe': {'a': [-1], 'b': 2, 'c': -10}}
    >>> c.hehe.pop("a")
    [-1]
    >>> c
    {'hehe': {'b': 2, 'c': -10}}

    """

    __name__ = "CFG"

    def __init__(self, *args, **kwargs) -> None:
        if len(args) > 1:
            raise TypeError(f"expected at most 1 arguments, got {len(args)}")
        elif len(args) == 1:
            d = args[0]
            assert isinstance(d, MutableMapping)
        else:
            d = {}
        if kwargs:
            d.update(**kwargs)
        for k, v in d.items():
            try:
                setattr(self, k, v)
            except Exception:
                dict.__setitem__(self, k, v)
        # Class attributes
        exclude_fields = ["update", "pop"]
        for k in self.__class__.__dict__:
            if not (k.startswith("__") and k.endswith("__")) and k not in exclude_fields:
                setattr(self, k, getattr(self, k))

    def __setattr__(self, name: str, value: Any) -> None:
        if isinstance(value, (list, tuple)):
            value = [self.__class__(x) if isinstance(x, dict) else x for x in value]
        elif isinstance(value, dict) and not isinstance(value, self.__class__):
            value = self.__class__(value)
        super().__setattr__(name, value)
        super().__setitem__(name, value)

    __setitem__ = __setattr__

    def update(self, new_cfg: Optional[MutableMapping] = None, **kwargs: Any) -> None:
        """The new hierarchical update method.

        Parameters
        ----------
        new_cfg : MutableMapping, optional
            The new configuration, by default None.
        **kwargs : dict, optional
            Key value pairs, by default None.

        """
        _new_cfg = new_cfg or CFG()
        if len(kwargs) > 0:  # avoid RecursionError
            _new_cfg.update(kwargs)
        for k in _new_cfg:
            if isinstance(_new_cfg[k], MutableMapping) and k in self:
                if isinstance(self[k], MutableMapping):
                    self[k].update(_new_cfg[k])
                else:  # for example, self[k] is `None`
                    self[k] = _new_cfg[k]  # deepcopy?
            else:
                try:
                    setattr(self, k, _new_cfg[k])
                except Exception:
                    dict.__setitem__(self, k, _new_cfg[k])

    def pop(self, key: str, default: Optional[Any] = None) -> Any:
        """The updated pop method.

        Parameters
        ----------
        key : str
            The key to pop.
        default : Any, optional
            The default value, by default None.

        """
        if key in self:
            delattr(self, key)
        return super().pop(key, default)

    def __repr__(self) -> str:
        return repr({k: v for k, v in self.items() if not callable(v)})

    def __str__(self) -> str:
        return str({k: v for k, v in self.items() if not callable(v)})


@dataclass
class DTYPE:
    """
    A dataclass to store the dtype information.

    Attributes
    ----------
    STR : str
        The string representation of the dtype.
    NP : np.dtype
        The numpy dtype.
    TORCH : torch.dtype
        The torch dtype.
    INT : int
        The int representation of the dtype, mainly used for `wfdb.rdrecord`.

    Examples
    --------
    >>> dtype = DTYPE("int16")
    >>> dtype
    DTYPE(STR='int16', NP=dtype('int16'), TORCH=torch.int16, INT=16)

    """

    STR: str
    NP: np.dtype = None
    TORCH: torch.dtype = None
    INT: int = None  # int representation of the dtype, mainly used for `wfdb.rdrecord`

    def __post_init__(self) -> None:
        """check consistency"""
        if self.NP is None:
            self.NP = np.dtype(self.STR)
        if self.TORCH is None:
            self.TORCH = eval(f"torch.{self.STR}")
        if self.INT is None:
            self.INT = int(re.search("\\d+", self.STR).group(0))
        assert all(
            [
                self.NP == getattr(np, self.STR),
                self.TORCH == getattr(torch, self.STR),
                self.INT == int(re.search("\\d+", self.STR).group(0)),
            ]
        ), "inconsistent dtype"

    @property
    def PRECISION(self) -> int:
        return self.INT


FLOAT16 = DTYPE("float16")
FLOAT32 = DTYPE("float32")
FLOAT64 = DTYPE("float64")
INT8 = DTYPE("int8")
INT16 = DTYPE("int16")
INT32 = DTYPE("int32")
INT64 = DTYPE("int64")


DEFAULTS = CFG()

DEFAULTS.log_dir = _PACKAGE_CACHE / "log"
DEFAULTS.checkpoints = _PACKAGE_CACHE / "checkpoints"
DEFAULTS.model_dir = _PACKAGE_CACHE / "saved_models"
DEFAULTS.working_dir = _PACKAGE_CACHE / "working_dir"
DEFAULTS.prefix = "TorchECG"

DEFAULTS.DTYPE = FLOAT32
DEFAULTS.DTYPE.TORCH = torch.float32  # torch.float64, torch.float16
DEFAULTS.str_dtype = str(DEFAULTS.DTYPE.TORCH).replace("torch.", "")
DEFAULTS.np_dtype = np.dtype(DEFAULTS.str_dtype)
DEFAULTS.dtype = DEFAULTS.DTYPE.TORCH

DEFAULTS.device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

DEFAULTS.eps = 1e-7

DEFAULTS.SEED = 42
DEFAULTS.RNG = np.random.default_rng(seed=DEFAULTS.SEED)
DEFAULTS.RNG_sample = partial(DEFAULTS.RNG.choice, replace=False, shuffle=False)
DEFAULTS.RNG_randint = partial(DEFAULTS.RNG.integers, endpoint=True)


def set_seed(seed: int) -> None:
    """
    Set the seed of the random number generator.

    Parameters
    ----------
    seed : int
        The seed to be set.

    """

    DEFAULTS.SEED = seed
    DEFAULTS.RNG = np.random.default_rng(seed=seed)
    DEFAULTS.RNG_sample = partial(DEFAULTS.RNG.choice, replace=False, shuffle=False)
    DEFAULTS.RNG_randint = partial(DEFAULTS.RNG.integers, endpoint=True)

    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)


DEFAULTS.set_seed = set_seed


def change_dtype(dtype: Union[str, np.dtype, torch.dtype]) -> None:
    """
    Change the dtype of the defaults.

    Parameters
    ----------
    dtype: str or numpy.dtype or torch.dtype,
        The dtype to be set.

    """
    # fmt: off
    _dtypes = [
        "float32", "float64", "float16",
        "int32", "int64", "int16", "int8", "uint8", "long",
    ]
    # fmt: on
    if isinstance(dtype, torch.dtype):
        _dtype = str(dtype).replace("torch.", "")
    elif isinstance(dtype, np.dtype):
        _dtype = str(dtype)
    elif isinstance(dtype, str):
        _dtype = dtype
    else:  # for example, dtype=np.float64
        try:
            _dtype = dtype.__name__
        except AttributeError:
            raise TypeError(f"`dtype` must be a str or np.dtype or torch.dtype, got {type(dtype)}")
    assert _dtype in _dtypes, f"`dtype` must be one of {_dtypes}, got {_dtype}"
    DEFAULTS.DTYPE = DTYPE(_dtype)
    DEFAULTS.dtype = DEFAULTS.DTYPE.TORCH


DEFAULTS.change_dtype = change_dtype



"""
"""

import warnings
from copy import deepcopy
from typing import List, Optional, Sequence, Tuple, Union

import torch
import torch.nn.functional as F
from torch import Tensor

from ..cfg import CFG
from ..model_configs.ecg_seq_lab_net import ECG_SEQ_LAB_NET_CONFIG
from ..utils import add_docstring
from .ecg_crnn import ECG_CRNN, ECG_CRNN_v1

__all__ = [
    "ECG_SEQ_LAB_NET",
]


class ECG_SEQ_LAB_NET(ECG_CRNN):
    """SOTA model from CPSC2019 challenge.

    Sequence labeling nets, for wave delineation, QRS complex detection, etc.
    Proposed in [:footcite:ct:`cai2020rpeak_seq_lab_net`].

    pipeline
    --------
    (multi-scopic, etc.) cnn --> head ((bidi-lstm -->) "attention" --> seq linear) -> output

    Parameters
    ----------
    classes : List[str]
        List of the classes for sequence labeling.
    n_leads : int
        Number of leads (number of input channels).
    config : dict, optional
        Other hyper-parameters, including kernel sizes, etc.
        Refer to corresponding config file.


    .. footbibliography::

    """

    __name__ = "ECG_SEQ_LAB_NET"
    __DEFAULT_CONFIG__ = {"recover_length": False}
    __DEFAULT_CONFIG__.update(deepcopy(ECG_SEQ_LAB_NET_CONFIG))

    def __init__(self, classes: Sequence[str], n_leads: int, config: Optional[CFG] = None) -> None:
        _config = CFG(deepcopy(self.__DEFAULT_CONFIG__))
        if not config:
            warnings.warn("No config is provided, using default config.", RuntimeWarning)
        _config.update(deepcopy(config) or {})
        _config.global_pool = "none"
        super().__init__(classes, n_leads, _config)

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass.

        Parameters
        ----------
        input : torch.Tensor
            Input tensor,
            of shape ``(batch_size, channels, seq_len)``.

        Returns
        -------
        pred : torch.Tensor
            Output tensor,
            of shape ``(batch_size, seq_len, n_classes)``

        """
        batch_size, channels, seq_len = input.shape

        pred = super().forward(input)  # (batch_size, len, n_classes)

        if self.config.recover_length:
            pred = F.interpolate(
                pred.permute(0, 2, 1),
                size=seq_len,
                mode="linear",
                align_corners=True,
            ).permute(0, 2, 1)

        return pred

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the model.

        Parameters
        ----------
        seq_len : int, optional
            Length of the 1d input signal tensor.
        batch_size : int, optional
            Batch size of the input signal tensor.

        Returns
        -------
        output_shape : sequence
            The output shape of the model.

        """
        output_shape = super().compute_output_shape(seq_len, batch_size)
        if self.config.recover_length:
            output_shape = (batch_size, seq_len, output_shape[-1])
        return output_shape

    @classmethod
    def from_v1(
        cls, v1_ckpt: str, device: Optional[torch.device] = None, return_config: bool = False
    ) -> Union["ECG_SEQ_LAB_NET", Tuple["ECG_SEQ_LAB_NET", dict]]:
        """Convert the v1 model to the current version.

        Parameters
        ----------
        v1_ckpt : str
            Path to the v1 checkpoint file.
        device : torch.device, optional
            The device to load the model to.
            Defaults to "cuda" if available, otherwise "cpu".
        return_config : bool, default False
            Whether to return the config dict.

        Returns
        -------
        model : ECG_SEQ_LAB_NET
            The converted model.
        config : dict
            The config dict. (if `return_config` is `True`)

        """
        v1_model, train_config = ECG_SEQ_LAB_NET_v1.from_checkpoint(v1_ckpt, device=device)
        model = cls(classes=v1_model.classes, n_leads=v1_model.n_leads, config=v1_model.config)
        model = model.to(v1_model.device)
        model.cnn.load_state_dict(v1_model.cnn.state_dict())
        if model.rnn.__class__.__name__ != "Identity":
            model.rnn.load_state_dict(v1_model.rnn.state_dict())
        if model.attn.__class__.__name__ != "Identity":
            model.attn.load_state_dict(v1_model.attn.state_dict())
        model.clf.load_state_dict(v1_model.clf.state_dict())
        del v1_model
        if return_config:
            return model, train_config
        return model

    @property
    def doi(self) -> List[str]:
        return list(set(super().doi + ["10.1109/access.2020.2997473"]))


@add_docstring(ECG_SEQ_LAB_NET.__doc__)
class ECG_SEQ_LAB_NET_v1(ECG_CRNN_v1):
    __name__ = "ECG_SEQ_LAB_NET_v1"
    __DEFAULT_CONFIG__ = {"recover_length": False}
    __DEFAULT_CONFIG__.update(deepcopy(ECG_SEQ_LAB_NET_CONFIG))

    def __init__(self, classes: Sequence[str], n_leads: int, config: Optional[CFG] = None) -> None:
        _config = CFG(deepcopy(self.__DEFAULT_CONFIG__))
        if not config:
            warnings.warn("No config is provided, using default config.", RuntimeWarning)
        _config.update(deepcopy(config) or {})
        _config.global_pool = "none"
        super().__init__(classes, n_leads, _config)

    def extract_features(self, input: Tensor) -> Tensor:
        """Extract feature map before the dense (linear) classifying layer(s).

        Parameters
        ----------
        input : torch.Tensor
            Input tensor,
            of shape ``(batch_size, channels, seq_len)``.

        Returns
        -------
        features : torch.Tensor
            Feature map tensor,
            of shape ``(batch_size, seq_len, channels)``.

        """
        # cnn
        cnn_output = self.cnn(input)  # (batch_size, channels, seq_len)

        # rnn or none
        if self.rnn:
            rnn_output = cnn_output.permute(2, 0, 1)  # (seq_len, batch_size, channels)
            rnn_output = self.rnn(rnn_output)  # (seq_len, batch_size, channels)
            rnn_output = rnn_output.permute(1, 2, 0)  # (batch_size, channels, seq_len)
        else:
            rnn_output = cnn_output

        # attention
        if self.attn:
            features = self.attn(rnn_output)  # (batch_size, channels, seq_len)
        else:
            features = rnn_output
        # features = features.permute(0, 2, 1)  # (batch_size, seq_len, channels)
        return features

    @add_docstring(ECG_SEQ_LAB_NET.forward.__doc__)
    def forward(self, input: Tensor) -> Tensor:
        batch_size, channels, seq_len = input.shape

        pred = super().forward(input)  # (batch_size, len, n_classes)

        if self.config.recover_length:
            pred = F.interpolate(
                pred.permute(0, 2, 1),
                size=seq_len,
                mode="linear",
                align_corners=True,
            ).permute(0, 2, 1)

        return pred

    @property
    def doi(self) -> List[str]:
        return list(set(super().doi + ["10.1109/access.2020.2997473"]))



"""
BSD 3-Clause License

Copyright (c) 2024, The Alphanumerics Lab

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


Modified from https://github.com/alphanumericslab/ecg-image-kit/blob/main/codes/ecg-image-generator/ecg_plot.py
"""

import os
import random
from math import ceil
from random import randint
from typing import Dict, List, Literal, Optional, Tuple, Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import AutoMinorLocator

__all__ = ["ecg_plot"]


standard_values = {
    "y_grid_size": 0.5,
    "x_grid_size": 0.2,
    "y_grid_inch": 5 / 25.4,
    "x_grid_inch": 5 / 25.4,
    "grid_line_width": 0.5,
    "lead_name_offset": 0.5,
    "lead_fontsize": 11.0,
    "x_gap": 1.0,
    "y_gap": 0.5,
    "display_factor": 1.0,
    "line_width": 0.75,
    "row_height": 8.0,
    "dc_offset_length": 0.2,
    "lead_length": 3.0,
    "V1_length": 12.0,
    "width": 11.0,
    "height": 8.5,
}

standard_major_colors = {
    "colour1": (0.4274, 0.196, 0.1843),  # brown
    "colour2": (1, 0.796, 0.866),  # pink
    "colour3": (0.0, 0.0, 0.4),  # blue
    "colour4": (0, 0.3, 0.0),  # green
    "colour5": (1, 0, 0),  # red
}


standard_minor_colors = {
    "colour1": (0.5882, 0.4196, 0.3960),
    "colour2": (0.996, 0.9294, 0.9725),
    "colour3": (0.0, 0, 0.7),
    "colour4": (0, 0.8, 0.3),
    "colour5": (0.996, 0.8745, 0.8588),
}

papersize_values = {
    "A0": (33.1, 46.8),
    "A1": (33.1, 23.39),
    "A2": (16.54, 23.39),
    "A3": (11.69, 16.54),
    "A4": (8.27, 11.69),
    "letter": (8.5, 11),
}

leadNames_12 = ["III", "aVF", "V3", "V6", "II", "aVL", "V2", "V5", "I", "aVR", "V1", "V4"]


def inches_to_dots(value: float, resolution: int) -> float:
    return value * resolution


def create_signal_dictionary(signal: np.ndarray, full_leads: List[str]) -> Dict[str, np.ndarray]:
    record_dict = {}
    for k in range(len(full_leads)):
        record_dict[full_leads[k]] = signal[k]
    return record_dict


def ecg_plot(
    ecg: Dict[str, np.ndarray],
    sample_rate: int,
    columns: int,
    rec_file_name: Union[str, bytes, os.PathLike],
    output_dir: Union[str, bytes, os.PathLike],
    resolution: int = 200,
    pad_inches: int = 0,
    lead_index: Optional[List[str]] = None,
    full_mode: str = "None",
    store_text_bbox: bool = False,
    units: str = "mV",
    papersize: Optional[Literal["A0", "A1", "A2", "A3", "A4", "letter"]] = None,
    x_gap: float = standard_values["x_gap"],
    y_gap: float = standard_values["y_gap"],
    display_factor: float = standard_values["display_factor"],
    line_width: float = standard_values["line_width"],
    title: Optional[str] = None,
    style: Optional[Literal["bw", "color"]] = None,
    row_height: float = standard_values["row_height"],
    show_lead_name: bool = True,
    show_grid: bool = False,
    show_dc_pulse: bool = False,
    y_grid: Optional[float] = None,
    x_grid: Optional[float] = None,
    standard_colours: Literal[0, 1, 2, 3, 4, 5] = 0,
    bbox: bool = False,
    print_txt: bool = False,
    save_format: Optional[List[Literal["png", "pdf", "svg", "jpg"]]] = None,
) -> Tuple[float, float]:
    """Function to plot raw ECG signal.

    Parameters
    ----------
    ecg : Dict[str, np.ndarray]
        Dictionary of ECG signals with lead names as keys,
        values as 1D numpy arrays.
    sample_rate : int
        Sampling rate of the ECG signal.
    columns : int
        Number of columns to be plotted in each row.
    rec_file_name : `path-like`
        Name of the record file.
    output_dir : `path-like`
        Output directory.
    resolution : int, default 200
        Resolution of the output image.
    pad_inches : int, default ``0``
        Padding of white margin along the image in inches.
    lead_index : List[str], optional
        Order of lead indices to be plotted.
        By default, the order is the same as the order in ``ecg``.
    full_mode : str, default "None"
        Sets the lead to add at the bottom of the paper ECG as a long strip.
        If ``"None"``, no lead is added at the bottom.
        If not ``"None"``, the lead ``"full" + full_mode`` must be present in ``ecg``.
    store_text_bbox : bool, default False
        If ``True``, stores the bounding box of the text in a text file.
    units : str, default "mV"
        Units of the ECG signal. NOT used currently.
    papersize : {"A0", "A1", "A2", "A3", "A4", "letter"}, default None
        Size of the paper to plot the ECG on.
    x_gap : float, default 1.0
        Gap between paper x axis border and signal plot.
    y_gap : float, default 0.5
        Gap between paper y axis border and signal plot.
    display_factor : float, default 1.0
        Factor to scale the ECG signal by.
    line_width : float, default 0.75
        Width of line tracing the ECG.
    title : str, optional
        Title of the figure.
    style : {"bw", "color"}, optional
        Sets the style of the plot.
    row_height : float, default 8.0
        Gap between corresponding ECG rows.
    show_lead_name : bool, default True
        Option to show lead names or skip.
    show_grid : bool, default False
        Turn grid on or off.
    show_dc_pulse : bool, default False
        Option to show DC pulse.
    y_grid : float, optional
        Sets the y grid size in inches.
    x_grid : float, optional
        Sets the x grid size in inches.
    standard_colours : {0, 1, 2, 3, 4, 5}, default 0
        Sets the colour of the plot grid.
    bbox : bool, default False
        If ``True``, stores the bounding box of the lead in a text file.
    print_txt : bool, default False
        If ``True``, prints the metadata of the plot.
    save_format : {"png", "pdf", "svg", "jpg"}, optional
        Format to save the plot in.
        If ``None``, the plot is not saved.

    Returns
    -------
    x_grid_dots : float
        X grid size in dots.
    y_grid_dots : float
        Y grid size in dots.

    """

    matplotlib.use("Agg")
    randindex = randint(0, 99)
    random_sampler = random.uniform(-0.05, 0.004)

    # check if the ecg dict is empty
    if ecg == {}:
        return 0, 0

    secs = len(list(ecg.items())[0][1]) / sample_rate

    assert secs * columns <= 10, "Total time of ECG signal of one row cannot exceed 10 seconds"

    if lead_index is None:
        lead_index = list(ecg.keys())
    leads = len(lead_index)

    rows = int(ceil(leads / columns))

    if full_mode != "None":
        rows += 1
        leads += 1

    # Grid calibration
    # Each big grid corresponds to 0.2 seconds and 0.5 mV
    # To do: Select grid size in a better way
    y_grid_size = standard_values["y_grid_size"]
    x_grid_size = standard_values["x_grid_size"]
    grid_line_width = standard_values["grid_line_width"]
    lead_name_offset = standard_values["lead_name_offset"]
    lead_fontsize = standard_values["lead_fontsize"]

    # Set max and min coordinates to mark grid. Offset x_max slightly (i.e by 1 column width)

    if not papersize:
        width = standard_values["width"]
        height = standard_values["height"]
    else:
        width = papersize_values[papersize][1]
        height = papersize_values[papersize][0]

    if y_grid is None:
        y_grid = standard_values["y_grid_inch"]
    if x_grid is None:
        x_grid = standard_values["x_grid_inch"]
    y_grid_dots = y_grid * resolution
    x_grid_dots = x_grid * resolution

    # row_height = height * y_grid_size/(y_grid*(rows+2))
    row_height = (height * y_grid_size / y_grid) / (rows + 2)
    x_max = width * x_grid_size / x_grid
    x_min = 0
    x_gap = np.floor(((x_max - (columns * secs)) / 2) / 0.2) * 0.2
    y_min = 0
    y_max = height * y_grid_size / y_grid

    # Set figure and subplot sizes
    fig, ax = plt.subplots(figsize=(width, height))

    fig.subplots_adjust(hspace=0, wspace=0, left=0, right=1, bottom=0, top=1)

    fig.suptitle(title)

    # Mark grid based on whether we want black and white or colour

    if style == "bw":
        color_major = (0.4, 0.4, 0.4)
        color_minor = (0.75, 0.75, 0.75)
        color_line = (0, 0, 0)
    elif standard_colours > 0:
        random_colour_index = standard_colours
        color_major = standard_major_colors["colour" + str(random_colour_index)]
        color_minor = standard_minor_colors["colour" + str(random_colour_index)]
        randcolorindex_grey = randint(0, 24)
        grey_random_color = random.uniform(0, 0.2)
        color_line = (grey_random_color, grey_random_color, grey_random_color)
    else:
        randcolorindex_red = randint(0, 24)
        major_random_color_sampler_red = random.uniform(0, 0.8)
        randcolorindex_green = randint(0, 24)
        major_random_color_sampler_green = random.uniform(0, 0.5)
        randcolorindex_blue = randint(0, 24)
        major_random_color_sampler_blue = random.uniform(0, 0.5)

        randcolorindex_minor = randint(0, 24)
        minor_offset = random.uniform(0, 0.2)
        minor_random_color_sampler_red = major_random_color_sampler_red + minor_offset
        minor_random_color_sampler_green = random.uniform(0, 0.5) + minor_offset
        minor_random_color_sampler_blue = random.uniform(0, 0.5) + minor_offset

        randcolorindex_grey = randint(0, 24)
        grey_random_color = random.uniform(0, 0.2)
        color_major = (major_random_color_sampler_red, major_random_color_sampler_green, major_random_color_sampler_blue)
        color_minor = (minor_random_color_sampler_red, minor_random_color_sampler_green, minor_random_color_sampler_blue)

        color_line = (grey_random_color, grey_random_color, grey_random_color)

    # Set grid
    # Standard ecg has grid size of 0.5 mV and 0.2 seconds. Set ticks accordingly
    if show_grid:
        ax.set_xticks(np.arange(x_min, x_max, x_grid_size))
        ax.set_yticks(np.arange(y_min, y_max, y_grid_size))
        ax.minorticks_on()

        ax.xaxis.set_minor_locator(AutoMinorLocator(5))

        # set grid line style
        ax.grid(which="major", linestyle="-", linewidth=grid_line_width, color=color_major)

        ax.grid(which="minor", linestyle="-", linewidth=grid_line_width, color=color_minor)

    else:
        ax.grid(False)
    ax.set_ylim(y_min, y_max)
    ax.set_xlim(x_min, x_max)
    ax.tick_params(axis="x", colors="white")
    ax.tick_params(axis="y", colors="white")
    lead_num = 0

    # Step size will be number of seconds per sample i.e 1/sampling_rate
    step = 1.0 / sample_rate

    dc_offset = 0
    if show_dc_pulse:
        dc_offset = sample_rate * standard_values["dc_offset_length"] * step
    # Iterate through each lead in lead_index array.
    y_offset = row_height / 2
    x_offset = 0

    text_bbox = []
    lead_bbox = []

    for i in np.arange(len(lead_index)):
        if len(lead_index) == 12:
            leadName = leadNames_12[i]
        else:
            leadName = lead_index[i]
        # y_offset is computed by shifting by a certain offset based on i,
        # and also by row_height/2 to account for half the waveform below the axis
        if i % columns == 0:
            y_offset += row_height

        # x_offset will be distance by which we shift the plot in each iteration
        if columns > 1:
            x_offset = (i % columns) * secs

        else:
            x_offset = 0

        # Create dc pulse wave to plot at the beginning of plot. Dc pulse will be 0.2 seconds
        x_range = np.arange(0, sample_rate * standard_values["dc_offset_length"] * step + 4 * step, step)
        dc_pulse = np.ones(len(x_range))
        dc_pulse = np.concatenate(((0, 0), dc_pulse[2:-2], (0, 0)))

        # Print lead name at .5 ( or 5 mm distance) from plot
        if show_lead_name:
            t1 = ax.text(x_offset + x_gap, y_offset - lead_name_offset - 0.2, leadName, fontsize=lead_fontsize)

            if store_text_bbox:
                renderer1 = fig.canvas.get_renderer()
                transf = ax.transData.inverted()
                bb = t1.get_window_extent()
                x1 = bb.x0 * resolution / fig.dpi
                y1 = bb.y0 * resolution / fig.dpi
                x2 = bb.x1 * resolution / fig.dpi
                y2 = bb.y1 * resolution / fig.dpi
                text_bbox.append([x1, y1, x2, y2, leadName])

        # If we are plotting the first row-1 plots, we plot the dc pulse prior to adding the waveform
        if columns == 1 and i in np.arange(0, rows):
            if show_dc_pulse:
                # Plot dc pulse for 0.2 seconds with 2 trailing and leading zeros to get the pulse
                t1 = ax.plot(x_range + x_offset + x_gap, dc_pulse + y_offset, linewidth=line_width * 1.5, color=color_line)
                if bbox:
                    renderer1 = fig.canvas.get_renderer()
                    transf = ax.transData.inverted()
                    bb = t1[0].get_window_extent()
                    x1, y1 = bb.x0 * resolution / fig.dpi, bb.y0 * resolution / fig.dpi
                    x2, y2 = bb.x1 * resolution / fig.dpi, bb.y1 * resolution / fig.dpi

        elif columns == 4 and i == 0 or i == 4 or i == 8:
            if show_dc_pulse:
                # Plot dc pulse for 0.2 seconds with 2 trailing and leading zeros to get the pulse
                t1 = ax.plot(
                    np.arange(0, sample_rate * standard_values["dc_offset_length"] * step + 4 * step, step) + x_offset + x_gap,
                    dc_pulse + y_offset,
                    linewidth=line_width * 1.5,
                    color=color_line,
                )
                if bbox:
                    renderer1 = fig.canvas.get_renderer()
                    transf = ax.transData.inverted()
                    bb = t1[0].get_window_extent()
                    x1, y1 = bb.x0 * resolution / fig.dpi, bb.y0 * resolution / fig.dpi
                    x2, y2 = bb.x1 * resolution / fig.dpi, bb.y1 * resolution / fig.dpi

        t1 = ax.plot(
            np.arange(0, len(ecg[leadName]) * step, step) + x_offset + dc_offset + x_gap,
            ecg[leadName] + y_offset,
            linewidth=line_width,
            color=color_line,
        )
        if bbox:
            renderer1 = fig.canvas.get_renderer()
            transf = ax.transData.inverted()
            bb = t1[0].get_window_extent()
            if show_dc_pulse is False or (columns == 4 and (i != 0 and i != 4 and i != 8)):
                x1, y1 = bb.x0 * resolution / fig.dpi, bb.y0 * resolution / fig.dpi
                x2, y2 = bb.x1 * resolution / fig.dpi, bb.y1 * resolution / fig.dpi
            else:
                y1 = min(y1, bb.y0 * resolution / fig.dpi)
                y2 = max(y2, bb.y1 * resolution / fig.dpi)
                x2 = bb.x1 * resolution / fig.dpi

            lead_bbox.append([x1, y1, x2, y2, 0])

        start_ind = round((x_offset + dc_offset + x_gap) * x_grid_dots / x_grid_size)
        end_ind = round((x_offset + dc_offset + x_gap + len(ecg[leadName]) * step) * x_grid_dots / x_grid_size)

    # Plotting longest lead for 12 seconds
    if full_mode != "None":
        if show_lead_name:
            t1 = ax.text(x_gap, row_height / 2 - lead_name_offset, full_mode, fontsize=lead_fontsize)

            if store_text_bbox:
                renderer1 = fig.canvas.get_renderer()
                transf = ax.transData.inverted()
                bb = t1.get_window_extent(renderer=fig.canvas.renderer)
                x1 = bb.x0 * resolution / fig.dpi
                y1 = bb.y0 * resolution / fig.dpi
                x2 = bb.x1 * resolution / fig.dpi
                y2 = bb.y1 * resolution / fig.dpi
                text_bbox.append([x1, y1, x2, y2, full_mode])

        if show_dc_pulse:
            t1 = ax.plot(
                x_range + x_gap,
                dc_pulse + row_height / 2 - lead_name_offset + 0.8,
                linewidth=line_width * 1.5,
                color=color_line,
            )

            if bbox:
                renderer1 = fig.canvas.get_renderer()
                transf = ax.transData.inverted()
                bb = t1[0].get_window_extent()
                x1, y1 = bb.x0 * resolution / fig.dpi, bb.y0 * resolution / fig.dpi
                x2, y2 = bb.x1 * resolution / fig.dpi, bb.y1 * resolution / fig.dpi

        dc_full_lead_offset = 0
        if show_dc_pulse:
            dc_full_lead_offset = sample_rate * standard_values["dc_offset_length"] * step

        t1 = ax.plot(
            np.arange(0, len(ecg["full" + full_mode]) * step, step) + x_gap + dc_full_lead_offset,
            ecg["full" + full_mode] + row_height / 2 - lead_name_offset + 0.8,
            linewidth=line_width,
            color=color_line,
        )

        if bbox:
            renderer1 = fig.canvas.get_renderer()
            transf = ax.transData.inverted()
            bb = t1[0].get_window_extent()
            if show_dc_pulse is False:
                x1, y1 = bb.x0 * resolution / fig.dpi, bb.y0 * resolution / fig.dpi
                x2, y2 = bb.x1 * resolution / fig.dpi, bb.y1 * resolution / fig.dpi
            else:
                y1 = min(y1, bb.y0 * resolution / fig.dpi)
                y2 = max(y2, bb.y1 * resolution / fig.dpi)
                x2 = bb.x1 * resolution / fig.dpi

            lead_bbox.append([x1, y1, x2, y2, 1])

        start_ind = round((dc_full_lead_offset + x_gap) * x_grid_dots / x_grid_size)
        end_ind = round((dc_full_lead_offset + x_gap + len(ecg["full" + full_mode]) * step) * x_grid_dots / x_grid_size)

    head, tail = os.path.split(rec_file_name)
    rec_file_name = os.path.join(output_dir, tail)

    if print_txt:
        pass
        # x_offset = 0.05
        # y_offset = int(y_max)
        # printed_text, attributes, flag = generate_template(full_header_file)

        # if flag:
        #     for l in range(0, len(printed_text), 1):

        #         for j in printed_text[l]:
        #             curr_l = ''
        #             if j in attributes.keys():
        #                 curr_l += str(attributes[j])
        #             ax.text(x_offset, y_offset, curr_l, fontsize=lead_fontsize)
        #             x_offset += 3

        #         y_offset -= 0.5
        #         x_offset = 0.05
        # else:
        #     for line in printed_text:
        #         ax.text(x_offset, y_offset, line, fontsize=lead_fontsize)
        #         y_offset -= 0.5

    # change x and y res
    ax.text(2, 0.5, "25mm/s", fontsize=lead_fontsize)
    ax.text(4, 0.5, "10mm/mV", fontsize=lead_fontsize)

    if save_format is not None:
        save_format = f""".{save_format.strip(".")}""".lower()
        assert save_format in [
            ".png",
            ".pdf",
            ".svg",
            ".jpg",
        ], f"save_format must be one of '.png', '.pdf', '.svg', '.jpg', but got {save_format}"
    else:
        return x_grid_dots, y_grid_dots

    plt.savefig(os.path.join(output_dir, tail + save_format), dpi=resolution, bbox_inches="tight", pad_inches=pad_inches)
    plt.close(fig)
    plt.clf()
    plt.cla()

    # margins
    right = pad_inches * resolution
    left = pad_inches * resolution
    top = pad_inches * resolution
    bottom = pad_inches * resolution

    if store_text_bbox:
        if os.path.exists(os.path.join(output_dir, "text_bounding_box")) is False:
            os.mkdir(os.path.join(output_dir, "text_bounding_box"))

        with open(os.path.join(output_dir, "text_bounding_box", tail + ".txt"), "w") as f:
            for i, l in enumerate(text_bbox):
                if pad_inches != 0:
                    l[0] += left
                    l[2] += left
                    l[1] += top
                    l[3] += top

                for val in l[:4]:
                    f.write(str(val))
                    f.write(",")
                f.write(str(l[4]))
                f.write("\n")

    if bbox:
        if os.path.exists(os.path.join(output_dir, "lead_bounding_box")) is False:
            os.mkdir(os.path.join(output_dir, "lead_bounding_box"))
        with open(os.path.join(output_dir, "lead_bounding_box", tail + ".txt"), "w") as f:
            for i, l in enumerate(lead_bbox):
                if pad_inches != 0:
                    l[0] += left
                    l[2] += left
                    l[1] += top
                    l[3] += top

                for val in l[:4]:
                    f.write(str(val))
                    f.write(",")
                f.write(str(l[4]))
                f.write("\n")

    return x_grid_dots, y_grid_dots



import os
from ast import literal_eval
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Union

import numpy as np
import pandas as pd
import wfdb
from tqdm.auto import tqdm

from ...cfg import DEFAULTS
from ...utils.misc import add_docstring, get_record_list_recursive3
from ..base import DataBaseInfo, PhysioNetDataBase

__all__ = [
    "PTBXL",
]


_PTBXL_INFO = DataBaseInfo(
    title="""
    PTB-XL, a large publicly available electrocardiography dataset
    """,
    about="""
    1. The PTB-XL database [1]_ is a large database of 21799 clinical 12-lead ECGs from 18869 patients of 10 second length collected with devices from Schiller AG over the course of nearly seven years between October 1989 and June 1996.
    2. The raw waveform data of the PTB-XL database was annotated by up to two cardiologists, who assigned potentially multiple ECG statements to each recording which were converted into a standardized set of SCP-ECG statements (scp_codes).
    3. The PTB-XL database contains 71 different ECG statements conforming to the SCP-ECG standard, including diagnostic, form, and rhythm statements.
    4. The waveform files of the PTB-XL database are stored in WaveForm DataBase (WFDB) format with 16 bit precision at a resolution of 1μV/LSB and a sampling frequency of 500Hz. A downsampled versions of the waveform data at a sampling frequency of 100Hz is also provided.
    5. In the metadata file (ptbxl_database.csv), each record of the PTB-XL database is identified by a unique ecg_id. The corresponding patient is encoded via patient_id. The paths to the original record (500 Hz) and a downsampled version of the record (100 Hz) are stored in `filename_hr` and `filename_lr`. The `report` field contains the diagnostic statements assigned to the record by the cardiologists. The `scp_codes` field contains the SCP-ECG statements assigned to the record which are formed as a dictionary with entries of the form `statement: likelihood`, where likelihood is set to 0 if unknown).
    6. The PTB-XL database underwent a 10-fold train-test splits (stored in the `strat_fold` field of the metadata file) obtained via stratified sampling while respecting patient assignments, i.e. all records of a particular patient were assigned to the same fold. Records in fold 9 and 10 underwent at least one human evaluation and are therefore of a particularly high label quality. It is proposed to use folds 1-8 as training set, fold 9 as validation set and fold 10 as test set.
    """,
    usage=[
        "Classification of ECG images",
    ],
    note="""
    1. A new comprehensive feature database named PTB-XL+ [2]_ was created to supplement the PTB-XL database.
    """,
    issues="""
    """,
    references=[
        "https://physionet.org/content/ptb-xl/",
        "https://physionet.org/content/ptb-xl-plus/",
    ],
    doi=[
        "https://doi.org/10.1038/s41597-023-02153-8",  # PTB-XL+ paper
        "https://doi.org/10.1038/s41597-020-0495-6",  # PTB-XL paper
        "https://doi.org/10.13026/nqsf-pc74",  # PTB-XL+ physionet
        "https://doi.org/10.13026/6sec-a640",  # PTB-XL physionet
    ],
)


@add_docstring(_PTBXL_INFO.format_database_docstring(), mode="prepend")
class PTBXL(PhysioNetDataBase):
    """
    Parameters
    ----------
    db_dir : `path-like`
        Storage path of the database.
        If not specified, data will be fetched from Physionet.
    working_dir : `path-like`, optional
        Working directory, to store intermediate files and log files.
    verbose : int, default 1
        Level of logging verbosity.
    feature_db_dir : `path-like`, optional
        Whether to include the feature database (the `PTB-XL+` database).
    kwargs : dict, optional
        Auxilliary key word arguments.

    """

    __name__ = "PTBXL"
    __metadata_file__ = "ptbxl_database.csv"
    __scp_statements_file__ = "scp_statements.csv"
    __100Hz_dir__ = "records100"
    __500Hz_dir__ = "records500"

    def __init__(
        self,
        db_dir: Union[str, bytes, os.PathLike],
        working_dir: Optional[Union[str, bytes, os.PathLike]] = None,
        verbose: int = 1,
        feature_db_dir: Optional[Union[str, bytes, os.PathLike]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(db_name="ptb-xl", db_dir=db_dir, working_dir=working_dir, verbose=verbose, **kwargs)
        self.feature_db_dir = Path(feature_db_dir).resolve() if feature_db_dir is not None else None
        if self.feature_db_dir is not None:
            self._feature_reader = PTBXLPlus(db_dir=self.feature_db_dir, verbose=verbose)
        else:
            self._feature_reader = None
        if self.fs is None:
            self.fs = 500
        self.data_ext = "dat"
        self.header_ext = "hea"
        self.record_pattern = "[\\d]{5}_[lh]r"

        self._df_records = None
        self._df_metadata = None
        self._df_scp_statements = None
        self._all_records = None
        self._all_subjects = None
        self._ls_rec()

    def _ls_rec(self) -> None:
        """Find all records in the database directory
        and store them (path, metadata, etc.) in a dataframe.
        """
        # locate the true database directory using the metadata file
        try:
            metadata_file = list(self.db_dir.rglob(self.__metadata_file__))[0]
        except IndexError:
            # raise FileNotFoundError(f"metadata file {self.__metadata_file__} not found in {self.db_dir}")
            self.logger.info(
                f"metadata file {self.__metadata_file__} not found in {self.db_dir}. "
                "Download the database first using the `download` method."
            )
            self._df_records = pd.DataFrame()
            self._df_metadata = pd.DataFrame()
            self._df_scp_statements = pd.DataFrame()
            self._all_records = []
            self._all_subjects = []
            return
        self.db_dir = metadata_file.parent.resolve()
        assert (self.db_dir / self.__scp_statements_file__).exists(), f"scp_statements file not found in {self.db_dir}"

        # read metadata file and scp_statements file
        self._df_metadata = pd.read_csv(self.db_dir / self.__metadata_file__)
        self._df_metadata["ecg_id"] = self._df_metadata["ecg_id"].apply(lambda x: f"{x:05d}")
        self._df_metadata.set_index("ecg_id", inplace=True)
        self._df_metadata["patient_id"] = self._df_metadata["patient_id"].astype(int)
        self._df_scp_statements = pd.read_csv(self.db_dir / self.__scp_statements_file__, index_col=0)

        self._df_records = self._df_metadata.copy()
        if self.fs == 100:
            self._df_records["path"] = self._df_records["filename_lr"].apply(lambda x: self.db_dir / x)
        else:
            self._df_records["path"] = self._df_records["filename_hr"].apply(lambda x: self.db_dir / x)
        # keep only records that exist
        self._df_records = self._df_records[
            self._df_records["path"].apply(lambda x: x.with_suffix(f".{self.data_ext}").exists())
        ]
        if self._subsample is not None:
            size = min(
                len(self._df_records),
                max(1, int(round(self._subsample * len(self._df_records)))),
            )
            self.logger.debug(f"subsample `{size}` records from `{len(self._df_records)}`")
            self._df_records = self._df_records.sample(n=size, random_state=DEFAULTS.SEED, replace=False)

        self._all_records = self._df_records.index.tolist()
        self._all_subjects = self._df_records["patient_id"].unique().tolist()

    def reset_fs(self, fs: int) -> None:
        """Reset the default sampling frequency.

        Parameters
        ----------
        fs : int
            The new sampling frequency.

        """
        self.fs = fs
        self._ls_rec()

    def load_metadata(
        self, rec: Union[str, int], items: Optional[Union[str, List[str]]] = None
    ) -> Union[Dict[str, Union[str, int, float]], str, int, float]:
        """Load the metadata of a record.

        Parameters
        ----------
        rec : str or int
            The record name (ecg_id) or the index of the record.
        items : str or list of str, optional
            The items to load.

        Returns
        -------
        metadata : dict
            The metadata of the record.

        """
        if isinstance(rec, int):
            rec = self._all_records[rec]
        if items is None:
            return self._df_metadata.loc[rec].to_dict()
        if isinstance(items, str):
            metadata = self._df_metadata.loc[rec, items]
            if isinstance(metadata, np.generic):
                return metadata.item()
        return self._df_metadata.loc[rec, items].to_dict()

    def load_ann(self, rec: Union[str, int], with_interpretation: bool = False) -> Dict[str, Union[float, Dict[str, Any]]]:
        """Load the annotation (the "scp_codes" field) of a record.

        Parameters
        ----------
        rec : str or int
            The record name (ecg_id) or the index of the record.
        with_interpretation : bool, default False
            Whether to include the interpretation of the statement.

        Returns
        -------
        ann : dict
            The annotation of the record, of the form ``{statement: likelihood}``.
            If ``with_interpretation`` is ``True``, the form is
            ``{statement: {"likelihood": likelihood, ...}}``,
            where ``...`` are other information of the statement.

        """
        ann = literal_eval(self.load_metadata(rec)["scp_codes"])
        if with_interpretation:
            for statement, likelihood in ann.items():
                ann[statement] = {"likelihood": likelihood}
                ann[statement].update(self._df_scp_statements.loc[statement].to_dict())
        return ann

    @property
    def database_info(self) -> DataBaseInfo:
        return _PTBXL_INFO

    @property
    def default_train_val_test_split(self) -> Dict[str, List[str]]:
        return {
            "train": self._df_records[self._df_records["strat_fold"] < 9].index.tolist(),
            "val": self._df_records[self._df_records["strat_fold"] == 9].index.tolist(),
            "test": self._df_records[self._df_records["strat_fold"] == 10].index.tolist(),
        }

    @property
    def default_train_val_split(self) -> Dict[str, List[str]]:
        return {
            "train": self._df_records[self._df_records["strat_fold"] < 10].index.tolist(),
            "val": self._df_records[self._df_records["strat_fold"] == 10].index.tolist(),
        }


_PTBXL_PLUS_INFO = DataBaseInfo(
    title="""
    PTB-XL+, a comprehensive electrocardiographic feature dataset
    """,
    about="""
    1. This database [1]_ is a comprehensive feature dataset that supplements the PTB-XL database [2]_.
    2. The features were extracted via two commercial (the University of Glasgow ECG Analysis Program (Uni-G) [3]_, GE Healthcare's Marquette™ 12SL™ (12SL) [4]_) and one open-source algorithm (ECGDeli [5]_).
    3. There are also automatic diagnosis statements from one commercial ECG analysis algorithm.
    4. Features are given a tabular format in the csv files features/12sl_features.csv, features/unig_features.csv and features/ecgdeli_features.csv, with mappings to standardized LOINC IDs given in features/feature_description.csv.
    5. Median beats are provided in the median_beats/12sl and median_beats/unig directories.
    6. Fiducial points are provided in the fiducial_points/ecgdeli directory.
    7. Diagnostic statements are provided in the labels/12sl_statements.csv, labels/unig_statements.csv and labels/ecgdeli_statements.csv files. The subdirectory labels/mapping contains mappings from the 12SL and PTB-XL diagnostic statements to SNOMED codes.
    """,
    usage=[
        "Evaluation of ECG (classification) models",
    ],
    note="""
    """,
    issues="""
    1. The file labels/mapping/ptbxlToSNOMED.csv contains bad lines with inconsistent number of columns.
    """,
    references=[
        "https://physionet.org/content/ptb-xl-plus/",
        "https://physionet.org/content/ptb-xl/",
        "Macfarlane, P., Devine, B. & Clark, E. (2005). The university of glasgow (uni-g) ecg analysis program. In Computers in Cardiology, 451–454.",
        "GE Healthcare (2019). Marquette 12SL ECG Analysis Program: Physician’s Guide. General Electric Company. 2056246-002C.",
        "Pilia, N., Nagel, C., Lenis, G., Becker, S., Dössel, O., Loewe, A. (2021). ECGdeli - an open source ECG delineation toolbox for MATLAB. SoftwareX 13, 100639.",
    ],
    doi=[
        "https://doi.org/10.1038/s41597-023-02153-8",  # PTB-XL+ paper
        "https://doi.org/10.13026/nqsf-pc74",  # PTB-XL+ physionet
    ],
)


@add_docstring(_PTBXL_PLUS_INFO.format_database_docstring(), mode="prepend")
class PTBXLPlus(PhysioNetDataBase):
    """
    Parameters
    ----------
    db_dir : `path-like`
        Storage path of the database.
        If not specified, data will be fetched from Physionet.
    working_dir : `path-like`, optional
        Working directory, to store intermediate files and log files.
    verbose : int, default 1
        Level of logging verbosity.
    kwargs : dict, optional
        Auxilliary key word arguments.

    """

    __name__ = "PTBXLPlus"

    def __init__(
        self,
        db_dir: Union[str, bytes, os.PathLike],
        working_dir: Optional[Union[str, bytes, os.PathLike]] = None,
        verbose: int = 1,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            db_dir=db_dir,
            db_name="ptb-xl-plus",
            working_dir=working_dir,
            verbose=verbose,
            **kwargs,
        )
        self.fs = None
        self.data_ext = "dat"  # median beats
        self.ann_ext = "atr"  # fiducial points
        self._feature_tables = None
        self._label_tables = None
        self._label_mappings = None
        self._ls_rec()

    def _ls_rec(self) -> None:
        """Find all records in the database directory
        and store them (path, metadata, etc.) in a dataframe.
        """
        records_file = list(self.db_dir.rglob("RECORDS"))
        if len(records_file) == 0:
            self._df_records = pd.DataFrame()
            self._all_records = []

            return
        self.db_dir = records_file[0].parent.resolve()

        # features: csv files, exclude the files in the "old" subdirectory
        feature_files = list((self.db_dir / "features").glob("*.csv"))
        self._feature_tables = {f.stem: pd.read_csv(f).dropna(how="all", axis=0) for f in feature_files}
        # turn the "ecg_id" column into string of 5 digits
        for table_name, table in self._feature_tables.items():
            if "ecg_id" in table.columns:
                table["ecg_id"] = table["ecg_id"].apply(lambda x: f"{x:05d}")

        # labels: csv files: mainly label descriptions and mappings (from 12sl/ptbxl to SNOMED)
        label_files = list((self.db_dir / "labels").glob("*.csv"))
        self._label_tables = {f.stem: pd.read_csv(f).dropna(how="all", axis=0) for f in label_files}
        mapping_files = list((self.db_dir / "labels" / "mapping").glob("*.csv"))
        # self._label_mappings = {f.stem: pd.read_csv(f) for f in mapping_files}
        for f in mapping_files:
            try:
                self._label_tables[f.stem] = pd.read_csv(f)
            except pd.errors.ParserError:
                file_content = [line.split(",") for line in f.read_text().splitlines()]
                common_len = max([max([idx for idx, item in enumerate(line) if item.strip()] + [0]) for line in file_content])
                file_content = [line[: common_len + 1] for line in file_content]
                self._label_tables[f.stem] = pd.DataFrame(file_content[1:], columns=file_content[0])
            self._label_tables[f.stem] = self._label_tables[f.stem].dropna(how="all", axis=0)
        # turn the "ecg_id" column into string of 5 digits
        for table_name, table in self._label_tables.items():
            if "ecg_id" in table.columns:
                table["ecg_id"] = table["ecg_id"].apply(lambda x: f"{x:05d}")

        # median beats: dat files, including "12sl" and "unig"
        self._df_records["12sl_path"] = get_record_list_recursive3(
            db_dir=self.db_dir / "median_beats" / "12sl",
            rec_patterns=f"[\\d]+_medians\\.{self.data_ext}",
            relative=False,
        )
        self._df_records["ecg_id"] = self._df_records["12sl_path"].apply(lambda x: int(Path(x).stem.split("_")[0]))
        df_unig = pd.DataFrame()
        df_unig["unig_path"] = get_record_list_recursive3(
            db_dir=self.db_dir / "median_beats" / "unig",
            rec_patterns=f"[\\d]+_medians\\.{self.data_ext}",
            relative=False,
        )
        df_unig["ecg_id"] = df_unig["unig_path"].apply(lambda x: int(Path(x).stem.split("_")[0]))
        # merge the two dataframes on "ecg_id" and keep empty cells
        self._df_records = self._df_records.merge(df_unig, on="ecg_id", how="outer")
        del df_unig

        # fiducial points: atr files
        df_fiducial = pd.DataFrame()
        df_fiducial["fiducial_path"] = get_record_list_recursive3(
            db_dir=self.db_dir / "fiducial_points" / "ecgdeli",
            rec_patterns=f"[\\d]+_points_[\\w_]+\\.{self.ann_ext}",
            relative=False,
        )
        df_fiducial["ecg_id"] = df_fiducial["fiducial_path"].apply(lambda x: int(Path(x).stem.split("_")[0]))
        fiducial_point_files = df_fiducial.groupby("ecg_id")["fiducial_path"].apply(list).to_dict()
        df_fiducial["fiducial_point_files"] = df_fiducial["ecg_id"].apply(lambda x: fiducial_point_files.get(x, []))
        # drop duplicates on "ecg_id"
        df_fiducial.drop_duplicates(subset="ecg_id", keep="first", inplace=True)
        # merge the two dataframes on "ecg_id" and keep empty cells
        self._df_records = self._df_records.merge(df_fiducial, on="ecg_id", how="outer")
        del df_fiducial

        # turn the "ecg_id" column into string of 5 digits
        self._df_records["ecg_id"] = self._df_records["ecg_id"].apply(lambda x: f"{x:05d}")
        self._df_records.set_index("ecg_id", inplace=True)
        self._all_records = self._df_records.index.tolist()

        # Fix potential bugs in the database
        self._fix_bugs()

    def load_data(self, rec: Union[str, int], source: Literal["12sl", "unig"] = "12sl") -> np.ndarray:
        """Load the data of a record.

        Parameters
        ----------
        rec : str or int
            The record name (ecg_id) or the index of the record.
        source : {"12sl", "unig"}, default "12sl"
            The data source to load.

        Returns
        -------
        data : ndarray
            The data of the record.

        """
        if isinstance(rec, int):
            rec = self._all_records[rec]
        if source == "12sl":
            path = self._df_records.loc[rec, "12sl_path"]
        elif source == "unig":
            path = self._df_records.loc[rec, "unig_path"]
        else:
            raise ValueError(f"unknown data source: {source}")
        if path is None:
            print(f"record {rec} not found in {source}")
            return None
        return wfdb.rdrecord(path).p_signal

    @add_docstring(load_data.__doc__)
    def load_median_beats(self, rec: Union[str, int], source: str = "12sl") -> np.ndarray:
        """alias of `load_data`."""
        return self.load_data(rec, source)

    def load_ann(self, rec: Union[str, int], source: Literal["12sl", "ptbxl"] = "12sl") -> Dict[str, Any]:
        """Load the annotation (diagnostic statements) of a record.

        Parameters
        ----------
        rec : str or int
            The record name (ecg_id) or the index of the record.
        source : {"12sl", "ptbxl"}, default "12sl"
            The annotation source to load.

        Returns
        -------
        ann : dict
            The annotation of the record.

        """
        if isinstance(rec, int):
            rec = self._all_records[rec]
        assert f"{source}_statements" in self._label_tables, f"source {source} not found in label tables"
        df = self._label_tables[f"{source}_statements"]
        if rec not in df["ecg_id"].values:
            return {}
        ann = df[df["ecg_id"] == rec].iloc[0].to_dict()
        ann.pop("ecg_id")
        ann = {key: literal_eval(val) for key, val in ann.items()}
        return ann

    def load_features(self, rec: Union[str, int], source: Literal["12sl", "unig", "ecgdeli"] = "12sl") -> Dict[str, float]:
        """Load the features of a record.

        Parameters
        ----------
        rec : str or int
            The record name (ecg_id) or the index of the record.
        source : {"12sl", "unig", "ecgdeli"}, default "12sl"
            The feature source to load.

        Returns
        -------
        features : dict
            The features of the record.

        """
        if isinstance(rec, int):
            rec = self._all_records[rec]
        assert f"{source}_features" in self._feature_tables, f"source {source} not found in feature tables"
        df = self._feature_tables[f"{source}_features"]
        if rec not in df["ecg_id"].values:
            return {}
        features = df[df["ecg_id"] == rec].iloc[0].to_dict()
        features.pop("ecg_id")
        return features

    def load_fiducial_points(
        self, rec: Union[str, int], leads: Optional[Union[str, List[str]]] = None
    ) -> Union[Dict[str, list], Dict[str, Dict[str, list]]]:
        """Load the fiducial points of a record.

        Parameters
        ----------
        rec : str or int
            The record name (ecg_id) or the index of the record.
        leads : str or list of str, optional
            The leads to load.
            If is None, load all leads.

        Returns
        -------
        fiducial_points : dict
            The fiducial points of the record.

        """
        if isinstance(rec, int):
            rec = self._all_records[rec]
        fiducial_point_files = self._df_records.loc[rec, "fiducial_point_files"]
        if not fiducial_point_files:
            return {}
        if leads is not None and isinstance(leads, str):
            leads = [leads]
        fiducial_points = {}
        for file in fiducial_point_files:
            lead = Path(file).stem.split("_")[-1]
            if leads is not None and lead not in leads:
                continue
            ann = wfdb.rdann(file, extension=self.ann_ext)
            fiducial_points[lead] = {
                "indices": ann.sample.tolist(),
                "labels": ann.aux_note,
            }
        return fiducial_points

    @property
    def database_info(self) -> DataBaseInfo:
        return _PTBXL_PLUS_INFO

    def _fix_bugs(self):
        """Fix bugs in the database.

        See https://github.com/MIT-LCP/wfdb-python/issues/528
        """
        flag_file = self.db_dir / "median_beats" / "12sl" / ".fixed"
        if flag_file.exists() or self.version > "1.0.1":
            return

        # fix the bug in the 12sl median beats
        with tqdm(
            self._df_records["12sl_path"].dropna(),
            total=len(self._df_records["12sl_path"].dropna()),
            desc="Fixing bugs in 12sl median beats header files",
            unit="record",
            dynamic_ncols=True,
            mininterval=1.0,
            disable=(self.verbose < 1),
        ) as pbar:
            for path in pbar:
                header_file = Path(path).with_suffix(".hea")
                header_content = header_file.read_text()
                header_content = header_content.replace("ge_median_beats_wfdb/", "")
                header_file.write_text(header_content)

        flag_file.touch()



"""
EfficientNet.

References
----------
1. Tan, M., & Le, Q. V. (2019). Efficientnet: Rethinking model scaling for convolutional neural networks. arXiv preprint arXiv:1905.11946.
2. Tan, M., & Le, Q. V. (2021). Efficientnetv2: Smaller models and faster training. arXiv preprint arXiv:2104.00298.
3. https://github.com/google/automl

"""

from typing import List

from torch import nn

from ...models._nets import Conv_Bn_Activation, DownSample, GlobalContextBlock, NonLocalBlock, SEBlock  # noqa: F401
from ...utils import CitationMixin, SizeMixin

__all__ = [
    "EfficientNet",
]


class EfficientNet(nn.Module, SizeMixin, CitationMixin):
    """
    Reference
    ---------
    1. Tan, M., & Le, Q. V. (2019). Efficientnet: Rethinking model scaling for convolutional neural networks. arXiv preprint arXiv:1905.11946.
    2. https://github.com/lukemelas/EfficientNet-PyTorch/blob/master/efficientnet_pytorch/model.py
    3. https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/efficientnet.py
    3. https://github.com/google/automl

    """

    __name__ = "EfficientNet"

    def __init__(self, in_channels: int, **config) -> None:
        super().__init__()
        raise NotImplementedError

    def forward(self):
        raise NotImplementedError

    def compute_output_shape(self):
        raise NotImplementedError

    @property
    def doi(self) -> List[str]:
        return list(set(self.config.get("doi", []) + ["10.48550/ARXIV.1905.11946"]))


class EfficientNetV2(nn.Module, SizeMixin):
    """
    Reference
    ---------
    1. Tan, M., & Le, Q. V. (2021). Efficientnetv2: Smaller models and faster training. arXiv preprint arXiv:2104.00298.
    2. https://github.com/d-li14/efficientnetv2.pytorch/blob/main/effnetv2.py
    3. https://github.com/lukemelas/EfficientNet-PyTorch/blob/master/efficientnet_pytorch/model.py
    4. https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/efficientnet.py
    5. https://github.com/google/automl

    """

    __name__ = "EfficientNetV2"

    def __init__(self, in_channels: int, **config) -> None:
        super().__init__()
        raise NotImplementedError

    def forward(self):
        raise NotImplementedError

    def compute_output_shape(self):
        raise NotImplementedError
